/*
kNRC - Noise reduction filter for VirtualDub
Copyright (C) 2002 Antonio Foranna

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation.
	
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
		
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
			
The author can be contacted at:
kreel@interfree.it
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"
#include "registry.h"

/***********************************************************************************************/

int dnrRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int dnrStartProc(FilterActivation *fa, const FilterFunctions *ff);
int dnrEndProc(FilterActivation *fa, const FilterFunctions *ff);
int dnrInitProc(FilterActivation *fa, const FilterFunctions *ff);
int dnrConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void dnrStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *buf); 

//long dnrParamProc(FilterActivation *fa, const FilterFunctions *ff);
void dnrScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool dnrFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

/***********************************************************************************************/

// Misc
#define REGISTRY_PROGRAM_NAME "SOFTWARE\\4N\\kNRC"
#define PROGRAM_NAME "kNRC"

#define THRESHOLDTABLELEN 8
#define ABS_SIZE	512
#define ABS_SHIFT	256

// DNR
#define FREQTHRESHOLD		4

#define REG_USE_DNR			"Use DNR"
#define REG_MINT			"Min threshold"
#define REG_MAXT			"Max threshold"
#define MAXTHRESHOLD		32
#define DEF_MINTHRESHOLD	4
#define DEF_MAXTHRESHOLD	12
#define DEF_USE_DNR			1

// Crop
#define REG_USE_CROP	"Use crop"
#define REG_LEFT		"Left"
#define REG_RIGHT		"Right"
#define REG_TOP			"Top"
#define REG_BOTTOM		"Bottom"
#define REG_XSHIFT		"X shift"
#define REG_YSHIFT		"Y shift"
#define DEF_CROP		0
#define DEF_SHIFT		0
#define DEF_USE_CROP	1

// Colors
#define REG_R		"Channel R"
#define REG_G		"Channel G"
#define REG_B		"Channel B"
#define MAXCOLORS	255
#define DEF_COLORS	0

///////////////////////////////////////////////////////////////////////////

#define CREATE_BUF(buf,type,size) \
    if (!(buf = (type *) new type[size])) \
        return 1; \
    memset(buf, 0, sizeof(type)*(size));

#define DELETE_BUF(buf) \
    delete[] buf; \
    buf = NULL;

/***********************************************************************************************/

static BYTE	NoEvents=0;

typedef struct MyFilterData
{
// Crop
DWORD	size;	// size in bytes of frames
WORD	h,w;	// heigth and width
WORD	left,
		right,
		top,
		bottom;
short	xShift,
		yShift;
BYTE	R,G,B;
BYTE	Use_crop;
long	cur_shift,
		dst_shift,
		dst_width,
		dst_heigth,
		xGap;

// DNR
BYTE	*old_data;				// previous frame
BYTE	MinT,					// Min threshold value
		MaxT;					// Max threshold value
short	*DiffAcc;				// Differences accumulator
BYTE	pLogThreshold[256];		// Logaritmic table of Thresholds
BYTE	ABS[ABS_SIZE],			// abs LUT to speedup computation
		*pABS;					// Has to point to middle of ABS
BYTE	Use_DNR;

// Misc
BYTE	PreviewMode;
HWND	hdlg,					// handle of config dialog (4 SetThresholdTable)
		hmL,					// handle of IDC_SLIDER_MINT (4 WM_HSCROLL)
		hML;					// handle of IDC_SLIDER_MAXT (4 WM_HSCROLL)
IFilterPreview *ifp;
} MyFilterData;

///////////////////////////////////////////////////////////////////////////


ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)dnrScriptConfig, "Config", "0iiiiiiiiiiiii" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};

struct FilterDefinition filterDEF_USE_DNR = {

    NULL, NULL, NULL,						// next, prev, module
    "kNRC",									// name
    "Reduces changes across frames, crops and centers images.",		// desc

	"Antonio Foranna",						// maker
    NULL,									// private_data
    sizeof(MyFilterData),					// inst_data_size

    dnrInitProc,							// initProc
    NULL,									// deinitProc
    dnrRunProc,								// runProc
    NULL,									// paramProc
    dnrConfigProc,							// configProc
    dnrStringProc,							// stringProc
    dnrStartProc,							// startProc
    dnrEndProc,								// endProc

    &script_obj,							// script_obj
    dnrFssProc,								// fssProc

};

/***********************************************************************************************/

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_dnr;

/***********************************************************************************************/

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_dnr = ff->addFilter(fm, &filterDEF_USE_DNR, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_dnr);
}

/***********************************************************************************************/

void InitVars(const FilterActivation *fa)
{
MyFilterData *mfd=(MyFilterData *)fa->filter_data;
short	ImageWidth=(short)fa->src.w,
		ImageHeigth=(short)fa->src.h;
DWORD	Color=mfd->R<<16|mfd->G<<8|mfd->B;
DWORD	nop=mfd->size>>2;
DWORD	*dst=(DWORD *)fa->dst.data;
short	dst_left,
		dst_top;

	if(!Color) // Black
		memset(fa->dst.data,0,mfd->size);
	else
		while(nop--)
			*(dst++)=Color;

	mfd->w=(WORD)fa->src.w;
	mfd->h=(WORD)fa->src.h;
	if(mfd->Use_crop)	
	{
		mfd->cur_shift=ImageWidth*mfd->top+mfd->left;
		mfd->dst_width=ImageWidth-mfd->left-mfd->right;
		mfd->dst_heigth=ImageHeigth-mfd->top-mfd->bottom;

		dst_top=((mfd->top+mfd->bottom)>>1)+mfd->yShift; // -mfd->yShift because Rows are stored in bottom up order
		if(dst_top<=-ImageHeigth)
			dst_top=-ImageHeigth+1;
		if(dst_top>=ImageHeigth)
			dst_top=ImageHeigth-1;
		if(dst_top<0)
		{
			mfd->cur_shift-=ImageWidth*dst_top;
			mfd->dst_heigth+=dst_top;
			dst_top=0;
		}

		if(mfd->dst_heigth<0)
			mfd->dst_heigth=0;
		if(mfd->dst_heigth+dst_top>ImageHeigth)
			mfd->dst_heigth=ImageHeigth-dst_top;

		dst_left=((mfd->left+mfd->right)>>1)+mfd->xShift;
		if(dst_left<=-ImageWidth)
			dst_left=ImageWidth+1;
		if(dst_left>=ImageWidth)
			dst_left=ImageWidth-1;
		if(dst_left<0)
		{
			mfd->cur_shift-=dst_left;
			mfd->dst_width+=dst_left;
			dst_left=0;
		}

		if(mfd->cur_shift>(long)mfd->size)
			mfd->cur_shift=mfd->size;
		if(mfd->cur_shift<0)
			mfd->cur_shift=0;

		mfd->dst_shift=ImageWidth*dst_top+dst_left;
		if(mfd->cur_shift+mfd->dst_shift>(long)mfd->size)
			mfd->dst_shift=0;

		if(mfd->dst_width<0)
			mfd->dst_width=0;
		if(mfd->dst_width+dst_left>ImageWidth)
			mfd->dst_width=ImageWidth-dst_left;
		if(!mfd->dst_width)
			mfd->dst_heigth=0;
	}
	else
	{
		mfd->cur_shift=0;
		mfd->dst_shift=0;
		mfd->dst_width=ImageWidth;
		mfd->dst_heigth=ImageHeigth;
	}

	mfd->xGap=ImageWidth-mfd->dst_width;
}

int dnrRunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
MyFilterData *mfd=(MyFilterData *)fa->filter_data;

	if(mfd->PreviewMode)
		InitVars(fa);

BYTE	*abs=mfd->pABS;
BYTE	*pLogThreshold=mfd->pLogThreshold,
		tThreshold;
short	*DiffAcc=mfd->DiffAcc;
DWORD	*old=(DWORD *)mfd->old_data,
		*cur=(DWORD *)fa->src.data,
		*dst=(DWORD *)fa->dst.data;
BYTE	*bold,
		*bcur,
		*bdst;

// kCenter
long	yOffset=0,
		xOffset=0,
		dst_width=mfd->dst_width,
		dst_heigth=mfd->dst_heigth,
		xGap=mfd->xGap;

	cur+=mfd->cur_shift;
	dst+=mfd->dst_shift;

	if(!mfd->Use_DNR)
	{
	WORD	ImageWidth=(WORD)fa->src.w;
		while(yOffset<mfd->dst_heigth)
		{
			memcpy(dst,cur,mfd->dst_width<<2);

			cur+=ImageWidth;
			dst+=ImageWidth;
			yOffset++;
	    }
		return 0;
	}

	while(yOffset<dst_heigth)
	{
	    bold=(BYTE *)old;
	    bcur=(BYTE *)cur;
	    bdst=(BYTE *)dst;

		*bdst    =(*bold    +(WORD)*bcur)    >>1;
		*(bdst+1)=(*(bold+1)+(WORD)*(bcur+1))>>1;
		*(bdst+2)=(*(bold+2)+(WORD)*(bcur+2))>>1;

		*DiffAcc    +=*(abs+*bdst    -*bold);
		*(DiffAcc+1)+=*(abs+*(bdst+1)-*(bold+1));
		*(DiffAcc+2)+=*(abs+*(bdst+2)-*(bold+2));

//		tThreshold=pLogThreshold[(*(bcur+2)+*(bcur+1)+*bcur)/3];
		tThreshold=*(pLogThreshold+((*(bcur+2)*299L+*(bcur+1)*587L+*bcur*114L)/1000));
		if(*DiffAcc    >=tThreshold ||
		   *(DiffAcc+1)>=tThreshold ||
		   *(DiffAcc+2)>=tThreshold)
		{
		    *dst=*cur;
			*DiffAcc=*(DiffAcc+1)=*(DiffAcc+2)=0;
		}

		*old=*dst;

		old++;
		cur++;
		dst++;
		DiffAcc+=3;
		if(++xOffset>=dst_width)
		{
			xOffset=0;
			yOffset++;
			cur+=xGap;
			dst+=xGap;
		}

    }

    return 0;
}
/***********************************************************************************************/

void LoadReg(MyFilterData *mfd)
{
registryClass *registry = new registryClass();
	if(registry->openReg(REGISTRY_PROGRAM_NAME))
	{
		mfd->MinT=registry->getSetRegByte(REG_MINT, DEF_MINTHRESHOLD);
		mfd->MaxT=registry->getSetRegByte(REG_MAXT, DEF_MAXTHRESHOLD);
		mfd->Use_DNR=registry->getSetRegByte(REG_USE_DNR, DEF_USE_DNR);

		mfd->left=registry->getSetRegWord(REG_LEFT,DEF_CROP);
		mfd->right=registry->getSetRegWord(REG_RIGHT,DEF_CROP);
		mfd->top=registry->getSetRegWord(REG_TOP,DEF_CROP);
		mfd->bottom=registry->getSetRegWord(REG_BOTTOM,DEF_CROP);
		mfd->xShift=registry->getSetRegWord(REG_XSHIFT,DEF_SHIFT);
		mfd->yShift=registry->getSetRegWord(REG_YSHIFT,DEF_SHIFT);
		mfd->R=registry->getSetRegByte(REG_R,DEF_COLORS);
		mfd->G=registry->getSetRegByte(REG_G,DEF_COLORS);
		mfd->B=registry->getSetRegByte(REG_B,DEF_COLORS);
		mfd->Use_crop=registry->getSetRegByte(REG_USE_DNR, DEF_USE_CROP);
	}
	else
	{
		mfd->MinT=GetProfileInt(PROGRAM_NAME, REG_MINT, DEF_MINTHRESHOLD);
		mfd->MaxT=GetProfileInt(PROGRAM_NAME, REG_MAXT, DEF_MAXTHRESHOLD);
		mfd->Use_DNR=GetProfileInt(PROGRAM_NAME, REG_USE_DNR, DEF_USE_DNR);

		mfd->left=GetProfileInt(PROGRAM_NAME,REG_LEFT,DEF_CROP);
		mfd->right=GetProfileInt(PROGRAM_NAME,REG_RIGHT,DEF_CROP);
		mfd->top=GetProfileInt(PROGRAM_NAME,REG_TOP,DEF_CROP);
		mfd->bottom=GetProfileInt(PROGRAM_NAME,REG_BOTTOM,DEF_CROP);
		mfd->xShift=GetProfileInt(PROGRAM_NAME,REG_XSHIFT,DEF_SHIFT);
		mfd->yShift=GetProfileInt(PROGRAM_NAME,REG_YSHIFT,DEF_SHIFT);
		mfd->R=GetProfileInt(PROGRAM_NAME,REG_R,DEF_COLORS);
		mfd->G=GetProfileInt(PROGRAM_NAME,REG_G,DEF_COLORS);
		mfd->B=GetProfileInt(PROGRAM_NAME,REG_B,DEF_COLORS);
		mfd->Use_crop=GetProfileInt(PROGRAM_NAME, REG_USE_CROP, DEF_USE_CROP);
	}
	if(mfd->MinT==0)
		mfd->MaxT=1;
	delete registry;
}

void SaveReg(MyFilterData *mfd)
{
registryClass *registry = new registryClass();
	if(registry->openCreateReg(REGISTRY_PROGRAM_NAME))
	{
		registry->setRegByte(REG_MINT, mfd->MinT);
		registry->setRegByte(REG_MAXT, mfd->MaxT);
		registry->setRegByte(REG_USE_DNR, mfd->Use_DNR);

		registry->setRegWord( REG_LEFT, mfd->left);
		registry->setRegWord( REG_RIGHT, mfd->right);
		registry->setRegWord( REG_TOP, mfd->top);
		registry->setRegWord( REG_BOTTOM, mfd->bottom);
		registry->setRegWord( REG_XSHIFT, mfd->xShift);
		registry->setRegWord( REG_YSHIFT, mfd->yShift);
		registry->setRegByte( REG_R, mfd->R);
		registry->setRegByte( REG_G, mfd->G);
		registry->setRegByte( REG_B, mfd->B);
		registry->setRegByte(REG_USE_CROP, mfd->Use_crop);
	}
	else
	{
	char leveltext[4];
		sprintf(leveltext, "%d", mfd->MinT);
		WriteProfileString(PROGRAM_NAME, REG_MINT,leveltext);
		sprintf(leveltext, "%d", mfd->MaxT);
		WriteProfileString(PROGRAM_NAME, REG_MAXT,leveltext);
		sprintf(leveltext, "%d", mfd->Use_DNR);
		WriteProfileString(PROGRAM_NAME, REG_USE_DNR,leveltext);

		sprintf(leveltext, "%i", mfd->left);
		WriteProfileString(PROGRAM_NAME,REG_LEFT,leveltext);
		sprintf(leveltext, "%i", mfd->right);
		WriteProfileString(PROGRAM_NAME,REG_RIGHT,leveltext);
		sprintf(leveltext, "%i", mfd->top);
		WriteProfileString(PROGRAM_NAME,REG_TOP,leveltext);
		sprintf(leveltext, "%i", mfd->bottom);
		WriteProfileString(PROGRAM_NAME,REG_BOTTOM,leveltext);
		sprintf(leveltext, "%i", mfd->xShift);
		WriteProfileString(PROGRAM_NAME,REG_XSHIFT,leveltext);
		sprintf(leveltext, "%i", mfd->yShift);
		WriteProfileString(PROGRAM_NAME,REG_YSHIFT,leveltext);
		sprintf(leveltext, "%i", mfd->R);
		WriteProfileString(PROGRAM_NAME,REG_R,leveltext);
		sprintf(leveltext, "%i", mfd->G);
		WriteProfileString(PROGRAM_NAME,REG_G,leveltext);
		sprintf(leveltext, "%i", mfd->B);
		WriteProfileString(PROGRAM_NAME,REG_B,leveltext);
		sprintf(leveltext, "%d", mfd->Use_crop);
		WriteProfileString(PROGRAM_NAME, REG_USE_CROP,leveltext);
	}
	delete registry;
}

void SetThresholdTable(BYTE MINT, BYTE MAXT, HWND hdlg, int L1, int L2)
{
	if(!hdlg)
		return;
	
WORD	i;
WORD	len=MAXT-MINT;
float	step=(float)len/(THRESHOLDTABLELEN-1);
WORD	tLog[]={2,4,8,16,32,64,128,256};
char	buf[50+THRESHOLDTABLELEN*9+1],
		tmp[9+1]; // "xxx=yyy, "

	if(len<1 || !hdlg)
		return;

	*buf=0;
	for(i=0; i<THRESHOLDTABLELEN/2-1; i++)
	{
		sprintf(tmp,"<%d: %d, ",tLog[i],(BYTE)(MINT+step*i+0.5));
		strcat(buf,tmp);
	} 
	sprintf(tmp,"<%d: %d",tLog[i],(BYTE)(MINT+step*i+0.5));
	strcat(buf,tmp);
	SetDlgItemText(hdlg, L1, buf); 

	*buf=0;
	for(i=THRESHOLDTABLELEN/2; i<THRESHOLDTABLELEN-1; i++)
	{
		sprintf(tmp,"<%d: %d, ",tLog[i],(BYTE)(MINT+step*i+0.5));
		strcat(buf,tmp);
	} 
	sprintf(tmp,"<%d: %d",tLog[i],(BYTE)(MINT+step*i+0.5));
	strcat(buf,tmp);
	SetDlgItemText(hdlg, L2, buf); 
}

void SetLogThreshold(MyFilterData *mfd)
{
WORD i,j;
WORD tLog[]={2,4,8,16,32,64,128,256};
WORD len=mfd->MaxT-mfd->MinT;
float step=(float)len/(THRESHOLDTABLELEN-1);

 if(len<1)
  return;

 for(i=0, j=0; i<256; i++)
 {
  mfd->pLogThreshold[i]=(BYTE)(mfd->MinT+step*j+0.5); // 0.5 to round value
  if(i>=tLog[j])
   j++;
 }
}



int dnrInitProc(FilterActivation *fa, const FilterFunctions *ff)
{
MyFilterData *mfd = (MyFilterData *)fa->filter_data;
int i;
//	mfd->ifp=fa->ifp; look at dnrConfigProc

	LoadReg(mfd);

	SetThresholdTable(mfd->MinT,mfd->MaxT,0,0,0);
	SetLogThreshold(mfd);

	mfd->hdlg=NULL;

	mfd->pABS=mfd->ABS+ABS_SHIFT;
	for(i=-ABS_SHIFT+1; i<0; i++)
		mfd->pABS[i]=(BYTE)-i;
	for(i=0; i<ABS_SHIFT; i++)
		mfd->pABS[i]=(BYTE)i;

    return 0;
}
/***********************************************************************************************/

void UpSliders(HWND hdlg, MyFilterData *mfd);

int dnrStartProc(FilterActivation *fa, const FilterFunctions *ff)
{
MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    mfd->size = fa->src.Size ();
	if(!mfd->size)
		return 1;

// buffer for previous frame
	CREATE_BUF(mfd->old_data,BYTE,mfd->size)
// Differences accumulator
	CREATE_BUF(mfd->DiffAcc,short,(mfd->size/4)*3)

// kCenter
	InitVars(fa);

	if(mfd->PreviewMode)
		UpSliders(mfd->hdlg,mfd);

    return 0;
}
/***********************************************************************************************/

int dnrEndProc(FilterActivation *fa, const FilterFunctions *ff)
{
MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    DELETE_BUF(mfd->old_data)
    DELETE_BUF(mfd->DiffAcc)

    return 0;
}
/***********************************************************************************************/

void dnrStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    sprintf(str, " m=%d, M=%d, DNR=%d, l=%i, r=%i t=%i, b=%i, x=%i, y=%i, Crop=%d, R=%d, G=%d, B=%d",
				 mfd->MinT,
				 mfd->MaxT,
				 mfd->Use_DNR,

				 mfd->left,
				 mfd->right,
				 mfd->bottom,
				 mfd->top,
				 mfd->xShift,
				 mfd->yShift,
				 mfd->Use_crop,

				 mfd->R,
				 mfd->G,
				 mfd->B
				 );
}
/***********************************************************************************************/

void dnrScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->MinT=argv[0].asInt();
	mfd->MaxT=argv[1].asInt();
	mfd->Use_DNR=argv[2].asInt();

	mfd->left=argv[3].asInt();
	mfd->right=argv[4].asInt();
	mfd->top=argv[5].asInt();
	mfd->bottom=argv[6].asInt();
	mfd->xShift=argv[7].asInt();
	mfd->yShift=argv[8].asInt();
	mfd->Use_crop=argv[9].asInt();

	mfd->R=argv[10].asInt();
	mfd->G=argv[11].asInt();
	mfd->B=argv[12].asInt();
}
/***********************************************************************************************/

bool dnrFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	_snprintf(buf, buflen, "Config(%d, %d, %d, %i, %i, %i, %i, %i, %i, %d, %d, %d, %d)",
								mfd->MinT,
								mfd->MaxT,
								mfd->Use_DNR,

								mfd->left,
								mfd->right,
								mfd->top,
								mfd->bottom,
								mfd->xShift,
							    mfd->yShift,
								mfd->Use_crop,

								mfd->R,
								mfd->G,
								mfd->B
								);

	return true;
}
/***********************************************************************************************/

void InitSliderEdit(HWND hdlg, DWORD S, DWORD E, DWORD Freq, DWORD Max, DWORD Val)
{
char buf[10];
	if(!hdlg)
		return;
	sprintf(buf, "%i", Val);
	NoEvents=1;
	SetDlgItemText(hdlg, E, buf);
	SendDlgItemMessage(hdlg, S, TBM_SETTICFREQ, Freq, 0);
	SendDlgItemMessage(hdlg, S, TBM_SETRANGE, false, MAKELONG (0, Max));
	EnableWindow(GetDlgItem(hdlg, S), Max);
	SendDlgItemMessage(hdlg, S, TBM_SETPOS, true, Val);
	NoEvents=0;
}

void InitSlider(HWND hdlg, DWORD S, DWORD Max, DWORD Val)
{
	if(!hdlg)
		return;
	NoEvents=1;
	EnableWindow(GetDlgItem(hdlg, S), Max);
	SendDlgItemMessage(hdlg, S, TBM_SETRANGE, false, MAKELONG (0, Max));
    SendDlgItemMessage(hdlg, S, TBM_SETPOS, true, Val);
	NoEvents=0;
}

void UpSliderEdit(HWND hdlg, DWORD S, DWORD E, DWORD Val)
{
char buf[10];
	if(!hdlg)
		return;
	sprintf(buf, "%i", Val);
	NoEvents=1;
	SetDlgItemText(hdlg, E, buf);
    SendDlgItemMessage(hdlg, S, TBM_SETPOS, true, Val);
	NoEvents=0;
}

void UpSlider(HWND hdlg, DWORD S, DWORD Val)
{
	if(!hdlg)
		return;
	NoEvents=1;
    SendDlgItemMessage(hdlg, S, TBM_SETPOS, true, Val);
	NoEvents=0;
}

void UpEdit(HWND hdlg, DWORD E, DWORD Val)
{
char buf[10];
	if(!hdlg)
		return;
	sprintf(buf, "%i", Val);
	NoEvents=1;
	SetDlgItemText(hdlg, E, buf);
	NoEvents=0;
}

void UpControls(HWND hdlg, MyFilterData *mfd)
{
	InitSliderEdit(hdlg,IDC_SLIDER_MINT,IDC_E_MINT,FREQTHRESHOLD,MAXTHRESHOLD,mfd->MinT);
	InitSliderEdit(hdlg,IDC_SLIDER_MAXT,IDC_E_MAXT,FREQTHRESHOLD,MAXTHRESHOLD,mfd->MaxT);
	InitSliderEdit(hdlg,IDC_SLIDER_TOP,IDC_E_TOP,FREQTHRESHOLD,mfd->h,mfd->top);
	InitSliderEdit(hdlg,IDC_SLIDER_BOTTOM,IDC_E_BOTTOM,FREQTHRESHOLD,mfd->h,mfd->bottom);
	InitSliderEdit(hdlg,IDC_SLIDER_LEFT,IDC_E_LEFT,FREQTHRESHOLD,mfd->w,mfd->left);
	InitSliderEdit(hdlg,IDC_SLIDER_RIGHT,IDC_E_RIGHT,FREQTHRESHOLD,mfd->w,mfd->right);
	InitSlider(hdlg, IDC_SLIDER_X, mfd->w*2, mfd->xShift+mfd->w);
	UpEdit(hdlg,IDC_E_X,mfd->xShift);
	InitSlider(hdlg, IDC_SLIDER_Y, mfd->h*2, mfd->yShift+mfd->h);
	UpEdit(hdlg,IDC_E_Y,mfd->yShift);
	InitSliderEdit(hdlg,IDC_SLIDER_R,IDC_E_R,FREQTHRESHOLD,MAXCOLORS,mfd->R);
	InitSliderEdit(hdlg,IDC_SLIDER_G,IDC_E_G,FREQTHRESHOLD,MAXCOLORS,mfd->G);
	InitSliderEdit(hdlg,IDC_SLIDER_B,IDC_E_B,FREQTHRESHOLD,MAXCOLORS,mfd->B);
}

void UpSliders(HWND hdlg, MyFilterData *mfd)
{
	InitSliderEdit(hdlg,IDC_SLIDER_MINT,IDC_E_MINT,FREQTHRESHOLD,MAXTHRESHOLD,mfd->MinT);
	InitSliderEdit(hdlg,IDC_SLIDER_MAXT,IDC_E_MAXT,FREQTHRESHOLD,MAXTHRESHOLD,mfd->MaxT);
	InitSliderEdit(hdlg,IDC_SLIDER_TOP,IDC_E_TOP,FREQTHRESHOLD,mfd->h,mfd->top);
	InitSliderEdit(hdlg,IDC_SLIDER_BOTTOM,IDC_E_BOTTOM,FREQTHRESHOLD,mfd->h,mfd->bottom);
	InitSliderEdit(hdlg,IDC_SLIDER_LEFT,IDC_E_LEFT,FREQTHRESHOLD,mfd->w,mfd->left);
	InitSliderEdit(hdlg,IDC_SLIDER_RIGHT,IDC_E_RIGHT,FREQTHRESHOLD,mfd->w,mfd->right);
	InitSlider(hdlg, IDC_SLIDER_X, mfd->w*2, mfd->xShift+mfd->w);
	UpEdit(hdlg,IDC_E_X,mfd->xShift);
	InitSlider(hdlg, IDC_SLIDER_Y, mfd->h*2, mfd->yShift+mfd->h);
	UpEdit(hdlg,IDC_E_Y,mfd->yShift);
}



BOOL CALLBACK dnrConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg)
	{
        case WM_INITDIALOG:
            {
                SetWindowLong(hdlg, DWL_USER, lParam);
                mfd = (MyFilterData *)lParam;

				mfd->hdlg=hdlg;
				mfd->hmL=GetDlgItem(hdlg,IDC_SLIDER_MINT);
				mfd->hML=GetDlgItem(hdlg,IDC_SLIDER_MAXT);
				mfd->ifp->InitButton(GetDlgItem(hdlg,IDC_BTN_PREVIEW));
				mfd->PreviewMode=0;

				SetThresholdTable(mfd->MinT,mfd->MaxT,hdlg,IDC_L_THRESHOLD1,IDC_L_THRESHOLD2);

				UpControls(hdlg,mfd);
				CheckDlgButton(hdlg,IDC_CHK_DNR,mfd->Use_DNR);
				CheckDlgButton(hdlg,IDC_CHK_CROP,mfd->Use_crop);
			}
            break;
			// END WM_INITDIALOG

        case WM_HSCROLL:
			{
				if(NoEvents)
					break;

				switch(GetDlgCtrlID((HWND)lParam))
				{
				case IDC_SLIDER_MINT:
					mfd->MinT=(BYTE)SendDlgItemMessage(hdlg, IDC_SLIDER_MINT, TBM_GETPOS, 0, 0);
					if(mfd->MinT>=mfd->MaxT)
					{
						mfd->MinT=mfd->MaxT-1;
						UpSlider(hdlg,IDC_SLIDER_MINT,mfd->MinT);
					}
					UpEdit(hdlg,IDC_E_MINT,mfd->MinT);
					SetThresholdTable(mfd->MinT,mfd->MaxT,hdlg,IDC_L_THRESHOLD1,IDC_L_THRESHOLD2);
					break;
				case IDC_SLIDER_MAXT:
					mfd->MaxT=(BYTE)SendDlgItemMessage(hdlg, IDC_SLIDER_MAXT, TBM_GETPOS, 0, 0);
					if(mfd->MaxT<=mfd->MinT)
					{
						mfd->MaxT=mfd->MinT+1;
						UpSlider(hdlg,IDC_SLIDER_MAXT,mfd->MaxT);
					}
					UpEdit(hdlg,IDC_E_MAXT,mfd->MaxT);
					SetThresholdTable(mfd->MinT,mfd->MaxT,hdlg,IDC_L_THRESHOLD1,IDC_L_THRESHOLD2);
					break;

				case IDC_SLIDER_TOP:
					mfd->top=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_TOP, TBM_GETPOS, 0, 0);
					if(mfd->top+mfd->bottom>mfd->h)
					{
						mfd->top=mfd->h-mfd->bottom;
						UpSlider(hdlg,IDC_SLIDER_TOP,mfd->top);
					}
					UpEdit(hdlg,IDC_E_TOP,mfd->top);
					break;
				case IDC_SLIDER_BOTTOM:
					mfd->bottom=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_BOTTOM, TBM_GETPOS, 0, 0);
					if(mfd->top+mfd->bottom>mfd->h)
					{
						mfd->bottom=mfd->h-mfd->top;
						UpSlider(hdlg,IDC_SLIDER_BOTTOM,mfd->bottom);
					}
					UpEdit(hdlg,IDC_E_BOTTOM,mfd->bottom);
					break;
				case IDC_SLIDER_LEFT:
					mfd->left=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_LEFT, TBM_GETPOS, 0, 0);
					if(mfd->left+mfd->right>mfd->w)
					{
						mfd->left=mfd->w-mfd->right;
						UpSlider(hdlg,IDC_SLIDER_LEFT,mfd->left);
					}
					UpEdit(hdlg,IDC_E_LEFT,mfd->left);
					break;
				case IDC_SLIDER_RIGHT:
					mfd->right=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_RIGHT, TBM_GETPOS, 0, 0);
					if(mfd->left+mfd->right>mfd->w)
					{
						mfd->right=mfd->w-mfd->left;
						UpSlider(hdlg,IDC_SLIDER_RIGHT,mfd->right);
					}
					UpEdit(hdlg,IDC_E_RIGHT,mfd->right);
					break;
				case IDC_SLIDER_X:
					mfd->xShift=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_X, TBM_GETPOS, 0, 0)-mfd->w;
					if(mfd->xShift+((mfd->left+mfd->right)>>1)>mfd->w)
					{
						mfd->xShift=mfd->w-((mfd->left+mfd->right)>>1);
						UpSlider(hdlg,IDC_SLIDER_X,mfd->xShift+mfd->w);
					}
					else
						if(mfd->xShift+((mfd->left+mfd->right)>>1)+(mfd->w-mfd->right-mfd->left)<0)
						{
							mfd->xShift=-(((mfd->left+mfd->right)>>1)+(mfd->w-mfd->right-mfd->left));
							UpSlider(hdlg,IDC_SLIDER_X,mfd->xShift+mfd->w);
						}
					UpEdit(hdlg,IDC_E_X,mfd->xShift);
					break;
				case IDC_SLIDER_Y:
					mfd->yShift=(WORD)SendDlgItemMessage(hdlg, IDC_SLIDER_Y, TBM_GETPOS, 0, 0)-mfd->h;
					if(mfd->yShift+((mfd->top+mfd->bottom)>>1)>mfd->h)
					{
						mfd->yShift=mfd->h-((mfd->top+mfd->bottom)>>1);
						UpSlider(hdlg,IDC_SLIDER_Y,mfd->yShift+mfd->h);
					}
					else
						if(mfd->yShift+((mfd->top+mfd->bottom)>>1)+(mfd->h-mfd->bottom-mfd->top)<0)
						{
							mfd->yShift=-(((mfd->top+mfd->bottom)>>1)+(mfd->h-mfd->bottom-mfd->top));
							UpSlider(hdlg,IDC_SLIDER_Y,mfd->yShift+mfd->h);
						}
					UpEdit(hdlg,IDC_E_Y,mfd->yShift);
					break;

				case IDC_SLIDER_R:
					mfd->R=(BYTE)SendDlgItemMessage(hdlg, IDC_SLIDER_R, TBM_GETPOS, 0, 0);
					UpEdit(hdlg,IDC_E_R,mfd->R);
					break;
				case IDC_SLIDER_G:
					mfd->G=(BYTE)SendDlgItemMessage(hdlg, IDC_SLIDER_G, TBM_GETPOS, 0, 0);
					UpEdit(hdlg,IDC_E_G,mfd->G);
					break;
				case IDC_SLIDER_B:
					mfd->B=(BYTE)SendDlgItemMessage(hdlg, IDC_SLIDER_B, TBM_GETPOS, 0, 0);
					UpEdit(hdlg,IDC_E_B,mfd->B);
					break;
				}

				if(mfd->ifp->isPreviewEnabled())
					mfd->ifp->RedoFrame();
			}
			break;
			// END WM_HSCROLL

		case WM_COMMAND:
		{
			if(NoEvents)
				break;

	        switch(HIWORD(wParam))
			{
			case EN_KILLFOCUS:
				switch(LOWORD(wParam))
				{
				case IDC_E_MINT:
					mfd->MinT=GetDlgItemInt (hdlg, IDC_E_MINT,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_MINT,IDC_E_MINT,mfd->MinT);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_MAXT:
					mfd->MaxT=GetDlgItemInt (hdlg, IDC_E_MAXT,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_MAXT,IDC_E_MAXT,mfd->MaxT);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;

				case IDC_E_TOP:
					mfd->top=GetDlgItemInt (hdlg, IDC_E_TOP,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_TOP,IDC_E_TOP,mfd->top);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_BOTTOM:
					mfd->bottom=GetDlgItemInt (hdlg, IDC_E_BOTTOM,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_BOTTOM,IDC_E_BOTTOM,mfd->bottom);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_LEFT:
					mfd->left=GetDlgItemInt (hdlg, IDC_E_LEFT,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_LEFT,IDC_E_LEFT,mfd->left);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_RIGHT:
					mfd->right=GetDlgItemInt (hdlg, IDC_E_RIGHT,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_RIGHT,IDC_E_RIGHT,mfd->right);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_X:
					mfd->xShift=GetDlgItemInt (hdlg, IDC_E_X,0,true); 
					UpSlider(hdlg,IDC_SLIDER_X,mfd->xShift+mfd->w);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_Y:
					mfd->yShift=(int)GetDlgItemInt (hdlg, IDC_E_Y,0,true);
					UpSlider(hdlg,IDC_SLIDER_Y,mfd->yShift+mfd->h);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;

				case IDC_E_R:
					mfd->R=(int)GetDlgItemInt (hdlg, IDC_E_R,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_R,IDC_E_R,mfd->R);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_G:
					mfd->G=(int)GetDlgItemInt (hdlg, IDC_E_G,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_G,IDC_E_G,mfd->G);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				case IDC_E_B:
					mfd->B=(int)GetDlgItemInt (hdlg, IDC_E_B,0,true);
					UpSliderEdit(hdlg,IDC_SLIDER_B,IDC_E_B,mfd->B);
					if(mfd->ifp->isPreviewEnabled())
						mfd->ifp->RedoFrame();
					break;
				// END EN_KILLFOCUS
             }
			break;

			default:
				switch(LOWORD(wParam))
				{
					case IDOK:
						SaveReg(mfd);

						if(mfd->ifp->isPreviewEnabled())
							mfd->ifp->Close();
						mfd->PreviewMode=0;

						EndDialog(hdlg, 0);

						SetLogThreshold(mfd);
					break;

		            case IDCANCEL:
						if(mfd->ifp->isPreviewEnabled())
							mfd->ifp->Close();
						mfd->PreviewMode=0;

						EndDialog(hdlg, 1);

						LoadReg(mfd);
		                return FALSE;

					case IDC_BTN_RELOAD:
						{
							LoadReg(mfd);
							SetThresholdTable(mfd->MinT,mfd->MaxT,0,0,0);
							UpControls(hdlg,mfd);
							if(mfd->ifp->isPreviewEnabled())
								mfd->ifp->RedoFrame();
						}
						break;

					case IDC_BTN_RESET:
						{
							mfd->MinT = DEF_MINTHRESHOLD;
							mfd->MaxT = DEF_MAXTHRESHOLD;
							SetThresholdTable(mfd->MinT,mfd->MaxT,0,0,0);
	
							mfd->left = DEF_CROP;
							mfd->right = DEF_CROP;
							mfd->top = DEF_CROP;
							mfd->bottom = DEF_CROP;
							mfd->xShift = DEF_SHIFT;
							mfd->yShift = DEF_SHIFT;
							mfd->R = DEF_COLORS;
							mfd->G = DEF_COLORS;
							mfd->B = DEF_COLORS;
						
							UpControls(hdlg,mfd);
							if(mfd->ifp->isPreviewEnabled())
								mfd->ifp->RedoFrame();
						}
						break;

					case IDC_BTN_PREVIEW:
						mfd->PreviewMode=!mfd->PreviewMode;
						mfd->ifp->Toggle(hdlg);
						break;

					case IDC_CHK_CROP:
						mfd->Use_crop=IsDlgButtonChecked(hdlg, IDC_CHK_CROP);
						if(mfd->ifp->isPreviewEnabled())
							mfd->ifp->RedoFrame();
						break;

					case IDC_CHK_DNR:
						mfd->Use_DNR=IsDlgButtonChecked(hdlg, IDC_CHK_DNR);
						if(mfd->ifp->isPreviewEnabled())
							mfd->ifp->RedoFrame();
						break;
				}

			}
			// END switch(HIWORD(wParam))
		break;
		}
		// END WM_COMMAND
		
		default: return FALSE;
	}
	return TRUE;
}
//---------------------------------------------------------------------------

int dnrConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
MyFilterData *mfd = (MyFilterData *)fa->filter_data;
int retVal;
	mfd->ifp=fa->ifp;
    retVal=DialogBoxParam(fa->filter->module->hInstModule,
					        MAKEINTRESOURCE(IDD_DIALOG_CONFIG), hwnd,
							dnrConfigDlgProc, (LPARAM)fa->filter_data);

	mfd->hdlg=NULL;

	return retVal;
}
