/*	Subtitler - Sub Station Alpha v2.x/4.x subtitling plugin for VirtualDub
	Copyright (C) 2000-2002 Avery Lee, All Rights Reserved.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	-----------------

	main.cpp: Handles interface to user and to VirtualDub filter API.
*/

#include <stdlib.h>
#include <crtdbg.h>
#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <math.h>
#include <string.h>
#include <stdio.h>

//#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)

#include "resource.h"
#include "filter.h"
#include "DialogueList.h"
#include "RenderList.h"
#include "FontCache.h"
#include "List.h"
#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#define MAX_LINES (64)
#define MAX_RECTS (128)

int subtitler_init(FilterActivation *fa, const FilterFunctions *ff);
void subtitler_deinit(FilterActivation *fa, const FilterFunctions *ff);
int subtitler_run(const FilterActivation *fa, const FilterFunctions *ff);
long subtitler_param(FilterActivation *fa, const FilterFunctions *ff);
int subtitler_config(FilterActivation *fa, const FilterFunctions *ff, HWND hWnd);
void subtitler_string(const FilterActivation *fa, const FilterFunctions *ff, char *buf);
int subtitler_start(FilterActivation *fa, const FilterFunctions *ff);
int subtitler_end(FilterActivation *fa, const FilterFunctions *ff);
void subtitler_script_config(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool subtitler_script_line(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

bool ParseSSAScript(DialogueList *pdl, char *szFile, int xresin, int yresin, RenderList *pRender);

static FilterDefinition *fd_subtitler;

struct SubtitlerRect {
	RECT r;
	long fore, back;
};

struct SubtitlerFilterData {
	IFilterPreview *ifp;
	DialogueList *pdlist;
	FontCache *pfc;
	char szScript[MAX_PATH];

	RenderList *pRender;
	int nRects, nRectMax;

	HDC hdcAntialias;

	long	lLastTimestamp;

	bool	fAntialias;
	bool	bAssumeGDISucks;
	int		nWrapMode;
};

ScriptFunctionDef subtitler_functions[]={
	{ (ScriptFunctionPtr)subtitler_script_config, "Config", "0is" },
	{ (ScriptFunctionPtr)subtitler_script_config, NULL, "0isii" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, subtitler_functions
};

static struct FilterDefinition filterDef_subtitler={
	0,0,NULL,
	"subtitler",
	"Adds subtitles to video from a Sub Station Alpha v2.x/4.x script.\n(Version 2.3)",
	"Avery Lee",
	NULL,sizeof(SubtitlerFilterData),
	subtitler_init,
	subtitler_deinit,
	subtitler_run,
	subtitler_param,
	subtitler_config,
	subtitler_string,
	subtitler_start,
	subtitler_end,
	&script_obj,
	subtitler_script_line,
};

////////////////////////////////////////////////////////////

void (*g_pThrowException)(const char *format, ...);
long (*g_pGetHostVerInfo)(char *buf, int len);

static void NullThrowException(const char *format, ...) {
}

static long NullGetHostVerInfo(char *buf, int len) {
	buf[0] = 0;
	return 0;
}

bool g_bUseMMX;

////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleCheckVersion();
extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

int __declspec(dllexport) __cdecl VirtualdubFilterModuleCheckVersion() {
	return VIRTUALDUB_FILTERDEF_VERSION;
}

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {

	fd_subtitler = ff->addFilter(fm, &filterDef_subtitler, sizeof(FilterDefinition));

	g_pThrowException = NullThrowException;
	g_pGetHostVerInfo = NullGetHostVerInfo;

#if VIRTUALDUB_FILTERDEF_VERSION >= 6
	if (vdfd_ver >= 6)
		g_pThrowException = ff->Except;
#endif

#if VIRTUALDUB_FILTERDEF_VERSION >= 7
	if (vdfd_ver >= 7)
		g_pGetHostVerInfo = ff->getHostVersionInfo;
#endif

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_subtitler);
}

////////////////////////////////////////////////////////////

int subtitler_init(FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;

	sfd->fAntialias = true;

	sfd->bAssumeGDISucks = (LONG)GetVersion() < 0;

	sfd->pdlist = new DialogueList();
	if (!sfd->pdlist)
		return 1;

	return 0;
}

void subtitler_deinit(FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;

	delete sfd->pdlist;
	sfd->pdlist = NULL;

	_CrtDumpMemoryLeaks();
}

long subtitler_param(FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;

	if (!sfd->fAntialias)
		fa->src.dwFlags = VFBitmap::NEEDS_HDC;

	fa->dst.offset	= fa->src.offset;

	return 0;
}

void subtitler_string(const FilterActivation *fa, const FilterFunctions *ff, char *buf) {
	SubtitlerFilterData *sfd;

	if (!(sfd = (SubtitlerFilterData *)fa->filter_data)) return;

	wsprintf(buf," (script: %s)", sfd->szScript);
}

/////////////////////////////////////////////////////////////

void SubtitlerConfigDlgBrowse(HWND hWnd) {
	OPENFILENAME ofn;
	char szFile[MAX_PATH];

	SendMessage(GetDlgItem(hWnd, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof szFile, (LPARAM)szFile);

	ofn.lStructSize			= sizeof(OPENFILENAME);
	ofn.hwndOwner			= hWnd;
	ofn.lpstrFilter			= "Sub Station Alpha v2.x/4.x script (*.ssa)\0*.ssa\0All files (*.*)\0*.*\0";
	ofn.lpstrCustomFilter	= NULL;
	ofn.nFilterIndex		= 1;
	ofn.lpstrFile			= szFile;
	ofn.nMaxFile			= sizeof szFile;
	ofn.lpstrFileTitle		= NULL;
	ofn.nMaxFileTitle		= 0;
	ofn.lpstrInitialDir		= NULL;
	ofn.lpstrTitle			= "Select subtitling script";
	ofn.Flags				= OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_ENABLESIZING;
	ofn.lpstrDefExt			= NULL;

	if (GetOpenFileName(&ofn)) {
		SendMessage(GetDlgItem(hWnd, IDC_SCRIPT_NAME), WM_SETTEXT, 0, (LPARAM)szFile);
	}
}

BOOL APIENTRY SubtitlerConfigDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)GetWindowLong(hDlg, DWL_USER);

    switch (message)
    {
        case WM_INITDIALOG:
			SetWindowLong(hDlg, DWL_USER, lParam);
			sfd = (SubtitlerFilterData *)lParam;

			SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_SETTEXT, 0, (LPARAM)sfd->szScript);
			CheckDlgButton(hDlg, IDC_ANTIALIAS, sfd->fAntialias ? BST_CHECKED:BST_UNCHECKED);
			CheckDlgButton(hDlg, IDC_GDISUCKS, sfd->bAssumeGDISucks ? BST_CHECKED:BST_UNCHECKED);

			switch(sfd->nWrapMode) {
			case 2:
				CheckDlgButton(hDlg, IDC_WRAP_MANUAL, BST_CHECKED);
				break;
			case 1:
				CheckDlgButton(hDlg, IDC_WRAP_AUTOMATIC, BST_CHECKED);
				break;
			case 0:
				CheckDlgButton(hDlg, IDC_WRAP_SMART, BST_CHECKED);
				break;
			}

			sfd->ifp->InitButton(GetDlgItem(hDlg, IDC_PREVIEW));

            return (TRUE);

        case WM_COMMAND:                 
			switch(LOWORD(wParam)) {

            case IDOK:
				{
					SubtitlerFilterData *sfd = (SubtitlerFilterData *)GetWindowLong(hDlg, DWL_USER);

					SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
				}
				EndDialog(hDlg, 0);
				return TRUE;

			case IDCANCEL:
                EndDialog(hDlg, 1);
                return TRUE;

			case IDC_BROWSE:
				SubtitlerConfigDlgBrowse(hDlg);
				return TRUE;

			case IDC_WRAP_MANUAL:
				if (IsDlgButtonChecked(hDlg, IDC_WRAP_MANUAL)) {
					if (sfd->nWrapMode != 2) {
						sfd->nWrapMode = 2;
						SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
						sfd->ifp->RedoSystem();
					}
				}
				return TRUE;

			case IDC_WRAP_AUTOMATIC:
				if (IsDlgButtonChecked(hDlg, IDC_WRAP_AUTOMATIC)) {
					if (sfd->nWrapMode != 1) {
						sfd->nWrapMode = 1;
						SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
						sfd->ifp->RedoSystem();
					}
				}
				return TRUE;

			case IDC_WRAP_SMART:
				if (IsDlgButtonChecked(hDlg, IDC_WRAP_SMART)) {
					if (sfd->nWrapMode != 0) {
						sfd->nWrapMode = 0;
						SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
						sfd->ifp->RedoSystem();
					}
				}
				return TRUE;

			case IDC_ANTIALIAS:
				{
					bool f = !!IsDlgButtonChecked(hDlg, IDC_ANTIALIAS);

					if (f != sfd->fAntialias) {
						sfd->fAntialias = f;
						SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
						sfd->ifp->RedoSystem();
					}
				}
				return TRUE;

			case IDC_GDISUCKS:
				{
					bool f = !!IsDlgButtonChecked(hDlg, IDC_GDISUCKS);

					if (f != sfd->bAssumeGDISucks) {
						sfd->bAssumeGDISucks = f;
						SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
						sfd->ifp->RedoSystem();
					}
				}
				return TRUE;

			case IDC_PREVIEW:
				SendMessage(GetDlgItem(hDlg, IDC_SCRIPT_NAME), WM_GETTEXT, sizeof sfd->szScript, (LPARAM)sfd->szScript);
				sfd->ifp->Toggle(hDlg);
				return TRUE;

            }
            break;
    }
    return FALSE;
}

int subtitler_config(FilterActivation *fa, const FilterFunctions *ff, HWND hWnd) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data, sfd_bak;

	sfd_bak = *sfd;
	sfd->ifp = fa->ifp;

	if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_SUBTITLER_CONFIG), hWnd, (DLGPROC)SubtitlerConfigDlgProc, (LPARAM)sfd)) {
		*sfd = sfd_bak;
		return 1;
	}

	return 0;
}

/////////////////////////////////////////////////////////////

int subtitler_start(FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;

	g_bUseMMX = ff->isMMXEnabled();

	sfd->pRender = new RenderList(128, 1.4, sfd->bAssumeGDISucks, (RenderList::eWrapMode)sfd->nWrapMode);

	if (!sfd->pRender)
		return 1;

	// If Antialiasing is enabled, create a display context for us to draw text
	// paths in.

	if (sfd->fAntialias) {
		sfd->hdcAntialias = CreateCompatibleDC(NULL);
		SetTextAlign(sfd->hdcAntialias, TA_TOP|TA_LEFT);
		SetTextColor(sfd->hdcAntialias, 0xffffff);
		SetBkColor(sfd->hdcAntialias, 0);
		SetBkMode(sfd->hdcAntialias, TRANSPARENT);
	}

	sfd->pfc = new FontCache(sfd->fAntialias ? sfd->hdcAntialias : fa->src.hdc, 8);
	if (!sfd->pfc)
		return 1;

	// Read in script.

	try {
		if (!ParseSSAScript(sfd->pdlist, sfd->szScript, fa->src.w, fa->src.h, sfd->pRender))
			return 1;
	} catch(const char *) {
		return 1;
	}

	if (!sfd->pdlist->Sort())
		return 1;

	sfd->pdlist->Start();

	sfd->lLastTimestamp = 0;

	return 0;
}

int subtitler_run(const FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;
	DialogueItem *pdi;

	// Bug in VirtualDub 1.3d causes fa->pfsi to be NULL.

	if (!sfd || !fa->pfsi) return 1;

	SetBkMode(fa->src.hdc, TRANSPARENT);
	SetTextAlign(fa->src.hdc, TA_TOP|TA_LEFT);

	sfd->nRects = 0;

//	wsprintf(buf, "Frame %ld", fa->pfsi->lCurrentSourceFrame);
//	doOutlineText(sfd, &fa->src, 10, 10, buf, -1, RGB(255,255,255), RGB(0,0,0));

	if (fa->pfsi->lSourceFrameMS < sfd->lLastTimestamp) {
		sfd->pRender->Clear();
		sfd->pdlist->Restart();
	}

	sfd->lLastTimestamp = fa->pfsi->lSourceFrameMS;

	_RPT1(0,"Timestamp: %d\n", fa->pfsi->lSourceFrameMS);

	sfd->pdlist->AdvanceClock2(fa->pfsi->lSourceFrameMS);

	while(pdi = sfd->pdlist->getNextExpired())
		sfd->pRender->RemoveItem(pdi);

	while(pdi = sfd->pdlist->getNextNew()) {
		sfd->pRender->AddItem(pdi, sfd->pfc, &fa->src, sfd->fAntialias ? sfd->hdcAntialias:fa->src.hdc, sfd->fAntialias);
	}

	if (sfd->fAntialias)
		sfd->pRender->RenderAntialiased(fa->pfsi->lSourceFrameMS, &fa->src, sfd->pfc, sfd->hdcAntialias);
	else
		sfd->pRender->RenderNormal(fa->pfsi->lSourceFrameMS, &fa->src, sfd->pfc);

	GdiFlush();

	_CrtCheckMemory();

	return 0;
}

int subtitler_end(FilterActivation *fa, const FilterFunctions *ff) {
	SubtitlerFilterData *sfd;

	if (!(sfd = (SubtitlerFilterData *)fa->filter_data)) return 1;

	sfd->pdlist->Stop();
	sfd->pdlist->Clear();

	delete sfd->pfc;
	sfd->pfc = NULL;

	delete sfd->pRender;
	sfd->pRender = NULL;

	if (sfd->fAntialias) {
		if (sfd->hdcAntialias) {
			DeleteDC(sfd->hdcAntialias);
			sfd->hdcAntialias = NULL;
		}
	}

	return 0;
}

void subtitler_script_config(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;

	sfd->fAntialias		= !!argv[0].asInt();
	strcpy(sfd->szScript, *argv[1].asString());

	sfd->nWrapMode = 0;

	if (argc>2) {
		sfd->nWrapMode = argv[2].asInt();
		if (sfd->nWrapMode < 0 || sfd->nWrapMode>=3)
			sfd->nWrapMode = 0;
		sfd->bAssumeGDISucks = !!argv[3].asInt();
	}
}

bool subtitler_script_line(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	SubtitlerFilterData *sfd = (SubtitlerFilterData *)fa->filter_data;
	char buf2[MAX_PATH];
	char c, *s, *t;

	t = sfd->szScript;
	s = buf2;

	do {
		c = *t++;
		if (c == '\\')
			c = '/';
	} while(*s++ = c);

	_snprintf(buf, buflen, "Config(%d, \"%s\", %d, %d)",
		sfd->fAntialias,
		buf2,
		sfd->nWrapMode,
		sfd->bAssumeGDISucks);

	return true;
}

/////////////////////////////////////////////////////////////

static char *pszParsePt;

static char *parsearg(char *str) {
	char *start, *s;

	if (!str)
		str = pszParsePt;

	while(isspace(*str))
		++str;

	start = str;
	while(*str && !strchr(",\n", *str))
		++str;

	if ((!*str || *str=='\n') && str==start)
		return NULL;

	s = str;

	if (*str)
		++str;

	while(s > start && isspace(s[-1]))
		--s;

	*s = 0;

	pszParsePt = str;

	return start;
}

static char *parseargfinish() {
	return pszParsePt;
}

long parsecolor(const char *pszText) {
	if (pszText[0] == '&' && toupper(pszText[1])=='H')
		return strtoul(pszText+2, NULL, 16);
	else
		return strtoul(pszText, NULL, 0);
}

// ParseSSAScript: Parses a Sub Station Alpha v2.x+ script

bool ParseSSAScript(DialogueList *pdl, char *szFile, int xresin, int yresin, RenderList *pRender) {
	static const DialogueStyleDefinition dfd_default={
		"*Default",
		"Arial",
		12,
		{ 0xFFFFFF, 0x80C080, 0xC0C000, },
		0x000000,
		ANSI_CHARSET,
		false,
		false,
		false,
		false,
		0,
		DALIGN_BOTTOM | DALIGN_CENTER,
		10,
		10,
		10,
		0,
	};

	FILE *f = NULL;
	bool fReturn = true;
	char buf[8192];
	DialogueStyleDefinition dfd;
	bool fUsingV4StyleFormat = false;
	int xres = xresin, yres = yresin;
	int dialogueno = 0;

	pdl->AddStyle(&dfd_default);

	try {
		if (!(f = fopen(szFile, "r")))
			throw strerror(errno);

		while(fgets(buf, sizeof buf, f)) {

			// lines starting with ; are comments

			if (buf[0] == ';')
				continue;

			// Look for Dialogue:, Style:, PlayResY:, and groups

			if (!strnicmp(buf, "[V4 Styles]", 11)) {
				fUsingV4StyleFormat = true;
			} else if (!strnicmp(buf, "PlayResY:", 9)) {
				const char *token;

				if (!(token = parsearg(buf + 9)))
					goto reject;

				yres = atoi(token);
				if (yres < 240)
					throw "Original subtitling Y resolution is unreasonably small";

				// 1280x1024 is 5:4, all other resolutions are 4:3.

				if (yres == 1024)
					xres = 1280;
				else
					xres = (yres*4)/3;

			} else if (!strnicmp(buf, "Style:", 6)) {
// Style: credits,Arial,22,8454143,8454143,8454143,2039583,-1,0,2,4,2,30,30,200
				const char *token;
				int i;

				if (!(token = parsearg(buf + 6))) goto reject;
				dfd.name = token;

				// rip off * from style name

				if (*dfd.name == '*')
					++dfd.name;

				if (!(token = parsearg(NULL))) goto reject;
				dfd.fontname = token;

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nHeight = atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.rgbColors[0] = parsecolor(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.rgbColors[1] = parsecolor(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.rgbColors[2] = parsecolor(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.rgbShadowColor = parsecolor(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.fBold = !!atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.fItalics = !!atoi(token);

				// GRR...

				if (fUsingV4StyleFormat) {
					// skip BorderStyle
					if (!(token = parsearg(NULL))) goto reject;
				}

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nBorderWidth = atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nShadowDepth = atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				i = atoi(token);

				dfd.daPos = i & 3;

				if (i & 4) dfd.daPos |= DALIGN_TOP;
				else if (i & 8) dfd.daPos |= DALIGN_MIDDLE;
				else dfd.daPos |= DALIGN_BOTTOM;

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nLeftMargin = atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nRightMargin = atoi(token);

				if (!(token = parsearg(NULL))) goto reject;
				dfd.nBottomMargin = atoi(token);

				// Not present in SSAv2 scripts

				dfd.nAlpha = 0;


				if (token = parsearg(NULL)) {
					dfd.nAlpha = atoi(token);		// Kotus defined bounds for this yet???

					if (token = parsearg(NULL)) {
						dfd.dwCharSet = (DWORD)atoi(token);
					}
				}

				dfd.fUnderline = false;

				pdl->AddStyle(&dfd);

				_RPT3(0,"Style defined: %s - font %s, %d point\n", dfd.name, dfd.fontname, dfd.nHeight);

			} else if (!strnicmp(buf, "Dialogue:", 9)) {
//	Dialogue: Marked=0,0:00:00.00,0:00:01.00,*Default,neighbor,0000,0000,0000,!Effect,It's on fire...

				int start_hr, start_min, start_sec, start_hs;
				int end_hr, end_min, end_sec, end_hs;
				char *s = buf + 9, *t;
				DialogueItem *pdi;
				char *token;
				const char *style;
				int nLeftMargin, nRightMargin, nBottomMargin;
				char colon1, colon2, colon3;

				// skip to first comma

				while(*s != ',')
					if (!*s++)
						goto reject;

				// read start, end time

				++s;

				token = parsearg(s);
				if (!token) goto reject;

				if (7 != sscanf(token, "%d%c%d%c%d%c%d", &start_hr, &colon1, &start_min, &colon2, &start_sec, &colon3, &start_hs)
					|| colon1==',' || colon2==',' || colon3==',')
					goto reject;

				token = parsearg(NULL);
				if (!token) goto reject;

				if (7 != sscanf(token, "%d%c%d%c%d%c%d", &end_hr, &colon1, &end_min, &colon2, &end_sec, &colon3, &end_hs)
					|| colon1==',' || colon2==',' || colon3==',')
					goto reject;

				// read style

				style = parsearg(NULL);
				if (!style)
					goto reject;

				// rip off leading * from the style name

				if (*style == '*')
					++style;

				// read name

				token = parsearg(NULL);

				// read left, right, bottom border override

				token = parsearg(NULL);
				nLeftMargin = atoi(token);

				token = parsearg(NULL);
				nRightMargin = atoi(token);

				token = parsearg(NULL);
				nBottomMargin = atoi(token);

				// read effect string

				int xscroll=0, yscroll=0, ytop=0, yheight=0;

				token = parsearg(NULL);

				if (!strnicmp(token, "Scroll up;", 10)) {
					int y1, y2, delay;

					if (3==sscanf(token+10, "%d;%d;%d", &delay, &y1, &y2)
						&& delay > 0) {

						// SSA allows top and bottom to be in any order.

						if (y1 > y2) {
							y1 ^= y2; y2 ^= y1; y1 ^= y2;
						}

						// Clip and scale.

						if (y1 < 0)
							y1 = 0;

						if (y2 > yres)
							y2 = yres;

						y1 = (y1 * yresin) / yres;
						y2 = (y2 * yresin) / yres;

						// Check for a null window.  If y1 == y2, replace
						// it with the full window.

						if (y1 >= y2) {
							y1 = 0;
							y2 = yresin;
						}

						// All done.

						ytop = y1;
						yheight = y2 - y1;

						yscroll = -MulDiv(yheight*256, delay, 100);
					}
				} else if (!strnicmp(token, "Banner;", 7)) {
					int delay;

					if (1==sscanf(token+7, "%d", &delay)
						&& delay > 0) {

						xscroll = -MulDiv(xresin*256, delay, 100);
					}
				}

				// read text

				token = s = parseargfinish();

				// parse text

				t = s;

				while(*t && *t != '\n')
					++t;

				*t = 0;

				// Don't add blank lines.

				if (*s) {
					t = strdup(s);

					if (!t)
						throw "[SSA v2 parser] out of memory";

					if (!(pdi = new DialogueItem)) {
						free(t);

						throw "[SSA v2 parser] out of memory";
					}

					pdi->order		= dialogueno++;
					pdi->text		= t;
					pdi->start_ms	= start_hr  * 3600000
									+ start_min *   60000
									+ start_sec *    1000
									+ start_hs  *      10;
					pdi->duration_ms=(end_hr  * 3600000
									+ end_min *   60000
									+ end_sec *    1000
									+ end_hs  *      10) - pdi->start_ms;
					pdi->nLeftMargin	= nLeftMargin;
					pdi->nRightMargin	= nRightMargin;
					pdi->nBottomMargin	= nBottomMargin;
					pdi->nXScrollRate	= xscroll;
					pdi->nYScrollRate	= yscroll;
					pdi->nYWindowPos	= ytop;
					pdi->nYWindowSize	= yheight;

					pdl->Add(pdi, style);

					_RPT3(0,"Added: %d+%d ms: [%s]\n", pdi->start_ms, pdi->duration_ms, pdi->text);
				}
			}
reject:
			;
		}

	} catch(int) {
		fReturn = false;
	} catch(const char *s) {
		if (g_pThrowException) {
			if (f)
				fclose(f);

			g_pThrowException(s);
		}
		MessageBox(NULL, s, "Error parsing Sub Station Alpha v2.x/4.x script", MB_OK);
	}

	pRender->SetRenderingScale(xresin, xres, yresin, yres);

	if (f)
		fclose(f);

	return fReturn;
}