//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <crtdbg.h>

#include "parser.h"
#include "compile.h"
#include "codegen.h"


///////////////////////////////////////////////////////////////////////////

class CodeHeap {
private:
	TreeTranslationNode *pFunctionList;
	TreeFunctionNode *pFunction;
	Instruction *pCode;
	int length, max;
	int sp;
	int sp_limit;

public:
	CodeHeap() : pCode(NULL), length(0), max(0), sp(0) {}
	~CodeHeap() {
		free(pCode);
	}

	void setFunctionList(TreeTranslationNode *list) {
		pFunctionList = list;
	}

	void setFunction(TreeFunctionNode *func) {
		pFunction = func;
	}

	TreeFunctionNode *getFunction() const {
		return pFunction;
	}

	Instruction *getCode() {
		Instruction *pCodeFinal;

		pCodeFinal = (Instruction *)realloc(pCode, length*sizeof(Instruction));

		if (pCodeFinal)
			pCode = NULL;

		return pCodeFinal;
	}

	int getIP() { return length; }

	Instruction& operator[](int i) {
		return pCode[i];
	}

	const Instruction& operator[](int i) const {
		return pCode[i];
	}

	int emit(InstructionType opcode, int r0=0, int r1=0, int r2=0) {
		if (length >= max) {

			if (!max)
				max = 512;
			else
				max <<= 1;

			pCode = (Instruction *)realloc(pCode, max*sizeof(Instruction));
		}

		if (!pCode)
			return 0;

		pCode[length].opcode = opcode;
		pCode[length].r0 = r0;
		pCode[length].r1 = r1;
		pCode[length].r2 = r2;

		return length++;
	}

	void setSP(int i) {
		sp = sp_limit = i;
	}

	int getSPLimit() {
		return sp_limit;
	}

	int push() {
		++sp;
		if (sp > sp_limit)
			sp_limit = sp;
		return sp-1;
	}

	int push(int i) {
		sp += i;
		if (sp > sp_limit)
			sp_limit = sp;
		return sp-i;
	}

	void pop() {
		--sp;
	}

	void pop(int i) {
		sp -= i;
	}

	TreeFunctionNode *getFunction(const char *fn) {
		TreeTranslationNode *unit;

		for(unit = pFunctionList; unit; unit=unit->next) {
			if (!strcmp(unit->func->name, fn))
				return unit->func;
		}

		return NULL;
	}
};

///////////////////////////////////////////////////////////////////////////
//
//	ctor/dtor
//
///////////////////////////////////////////////////////////////////////////

CodeGenerator::CodeGenerator(ParseNodeAllocator *_autofree, TreeTranslationNode *_pUnit)
: autofree(_autofree), pUnit(_pUnit) {
}

CodeGenerator::~CodeGenerator() {
}

///////////////////////////////////////////////////////////////////////////
//
//	basic code generation - lvalues
//
///////////////////////////////////////////////////////////////////////////

TreeVal TreeExpressionNode::compileRegValue(CodeHeap *output) {
	return TreeVal(-1, TYPE_VOID);
}

TreeVal TreeConstantNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "constants cannot be assigned a value");
}

TreeVal TreeVRefNode::compileLValue(CodeHeap *output, TreeVal v) {
	if (v.t != varnode->t) {
		if (varnode->t == TYPE_PIXEL64 && v.t == TYPE_PIXEL32) {
			output->emit(INST_PIXEL64TOPIXEL32, varnode->idx, v.r);
		} else
			throw ParseError(offset, "variable and expression types are incompatible");
	} else
		switch(v.t) {
		case TYPE_PIXEL64:	output->emit(INST_MOVEPIXEL64, varnode->idx, v.r);	break;
		case TYPE_PIXEL32:	output->emit(INST_MOVEPIXEL32, varnode->idx, v.r);	break;
		case TYPE_BITMAP:	output->emit(INST_MOVEBITMAP, varnode->idx, v.r);	break;
		case TYPE_INTEGER:	output->emit(INST_MOVEINT, varnode->idx, v.r);		break;
		}

	return v;
}

TreeVal TreeVRefNode::compileRegValue(CodeHeap *output) {
	return TreeVal(varnode->idx, varnode->t);
}

TreeVal TreeInNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "'in' cannot be assigned a value");
}

TreeVal TreeOutNode::compileLValue(CodeHeap *output, TreeVal v) {
	if (v.t == TYPE_PIXEL64)
		output->emit(INST_PIXEL64TOPIXEL32, scanner->activeidx, v.r);
	else if (v.t == TYPE_PIXEL32)
		output->emit(INST_MOVEPIXEL32, scanner->activeidx, v.r);
	else
		throw ParseError(offset, "'out' only accepts types 'pixel32' and 'pixel64'");

	return v;
}

TreeVal TreeOutNode::compileRegValue(CodeHeap *output) {
	return TreeVal(scanner->activeidx, TYPE_PIXEL32);
}

TreeVal TreeFunctionCallNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "function calls cannot be assigned a value");
}

TreeVal TreeIndexingNode::compileLValue(CodeHeap *output, TreeVal v) {
	TreeVal bv, xv, yv;

	bv = exp->compile(output);
	xv = xexp->compile(output);
	yv = yexp->compile(output);

	if (bv.t != TYPE_BITMAP)
		throw ParseError(offset, "only bitmaps can be indexed");

	if (xv.t != TYPE_INTEGER)
		throw ParseError(offset, "bitmap index must be an integer");

	if (yv.t != TYPE_INTEGER)
		throw ParseError(offset, "bitmap index must be an integer");

	output->emit(INST_SETPIXELBITMAP, bv.r, v.r, xv.r);

	output->pop(3);

	return v;
}

TreeVal TreeUnaryNode::compileLValue(CodeHeap *output, TreeVal v) {
	switch(op) {
	case OP_RED:
	case OP_GREEN:
	case OP_BLUE:
		{
			TreeVal x = exp->compileRegValue(output);

			if (x.r < 0)
				throw ParseError(offset, "channel insertion not possible on expression");

			if (v.t != TYPE_INTEGER)
				throw ParseError(offset, "only integers can be inserted into pixel channels");

			if (x.t == TYPE_PIXEL32) {
				switch(op) {
				case OP_RED:
					output->emit(INST_SETRPIXEL32, x.r, v.r);
					break;
				case OP_GREEN:
					output->emit(INST_SETGPIXEL32, x.r, v.r);
					break;
				case OP_BLUE:
					output->emit(INST_SETBPIXEL32, x.r, v.r);
					break;
				}
				v.t = TYPE_PIXEL32;
			} else if (x.t == TYPE_PIXEL64) {
				switch(op) {
				case OP_RED:
					output->emit(INST_SETRPIXEL64, x.r, v.r);
					break;
				case OP_GREEN:
					output->emit(INST_SETGPIXEL64, x.r, v.r);
					break;
				case OP_BLUE:
					output->emit(INST_SETBPIXEL64, x.r, v.r);
					break;
				}
			} else
				throw ParseError(offset, "channel insertion only valid for pixel32 and pixel64 types");

			v.t = TYPE_INTEGER;
		}
		break;
	default:
		throw ParseError(offset, "expression cannot be assigned a value");
	}

	return v;
}

TreeVal TreeBinaryNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "binary expressions cannot be assigned a value");
}

TreeVal TreeTernaryNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "ternary expressions cannot be assigned a value");
}

TreeVal TreeAssignNode::compileLValue(CodeHeap *output, TreeVal v) {
	throw ParseError(offset, "binary expressions cannot be assigned a value");
}

///////////////////////////////////////////////////////////////////////////
//
//	Basic code generation - rvalues.
//
///////////////////////////////////////////////////////////////////////////

TreeVal TreeConstantNode::compile(CodeHeap *output) {
	int r = output->push();

	if (type == TYPE_PIXEL32) {
		output->emit(INST_CONSTPIXEL32, r, val.pix);
		return TreeVal(r, TYPE_PIXEL32);
	} else {
		output->emit(INST_CONSTINT, r, val.i);
		return TreeVal(r, TYPE_INTEGER);
	}
}

TreeVal TreeVRefNode::compile(CodeHeap *output) {
	int r = output->push();

	switch(varnode->t) {
	case TYPE_BITMAP:
		output->emit(INST_MOVEBITMAP, r, varnode->idx);
		break;
	case TYPE_INTEGER:
		output->emit(INST_MOVEINT, r, varnode->idx);
		break;
	case TYPE_PIXEL32:
		output->emit(INST_MOVEPIXEL32, r, varnode->idx);
		break;
	case TYPE_PIXEL64:
		output->push();
		output->emit(INST_MOVEPIXEL64, r, varnode->idx);
		break;
	}

	return TreeVal(r, varnode->t);
}

TreeVal TreeInNode::compile(CodeHeap *output) {
	int r = output->push();

	output->emit(INST_MOVEPIXEL32, r, scanner->activeidx);

	return TreeVal(r, TYPE_PIXEL32);
}

TreeVal TreeOutNode::compile(CodeHeap *output) {
	throw ParseError(offset, "only assignments permitted to 'out' variable");
}

TreeVal TreeCallArgumentNode::compile(CodeHeap *output) {
	return exp->compile(output);
}

TreeVal TreeFunctionCallNode::compile(CodeHeap *output) {
	TreeCallArgumentNode *arg;
	TreeFunctionNode *func = output->getFunction(name);
	int fsize;

	if (!func)
		throw ParseError(offset, "function '%s' not defined", name);

	if (func->rettype == TYPE_PIXEL64)
		fsize = 2;
	else if (func->rettype != TYPE_VOID)
		fsize = 1;
	else
		fsize = 0;

	int rsp = output->push(fsize);

	for(arg = args; arg; arg=arg->nextarg) {
		TreeVal x = arg->compile(output);

		switch(x.t) {
		case TYPE_BITMAP:	output->emit(INST_PUSHBITMAP, 0, x.r, 0); break;
		case TYPE_INTEGER:	output->emit(INST_PUSHINT, 0, x.r, 0); break;
		case TYPE_PIXEL32:	output->emit(INST_PUSHPIXEL32, 0, x.r, 0); break;
		case TYPE_PIXEL64:	output->emit(INST_PUSHPIXEL64, 0, x.r, 0); break;
		}
	}

	output->emit(INST_CALL, 0, 0, (int)func);

	switch(func->rettype) {
	case TYPE_BITMAP:
		output->emit(INST_RVALBITMAP, rsp);
		break;
	case TYPE_INTEGER:
		output->emit(INST_RVALINT, rsp);
		break;
	case TYPE_PIXEL32:
		output->emit(INST_RVALPIXEL32, rsp);
		break;
	case TYPE_PIXEL64:
		output->emit(INST_RVALPIXEL64, rsp);
		break;
	}

	return TreeVal(rsp, func->rettype);
}

TreeVal TreeIndexingNode::compile(CodeHeap *output) {
	TreeVal bmp = exp->compile(output);
	TreeVal xr = xexp->compile(output);
	TreeVal yr = yexp->compile(output);

	if (bmp.t != TYPE_BITMAP)
		throw ParseError(offset, "only bitmaps can be indexed");

	if (xr.t != TYPE_INTEGER)
		throw ParseError(offset, "bitmap index must be an integer");

	if (yr.t != TYPE_INTEGER)
		throw ParseError(offset, "bitmap index must be an integer");

	output->emit(INST_GETPIXELBITMAP, bmp.r, xr.r, yr.r);

	output->pop(2);

	bmp.t = TYPE_PIXEL32;

	return bmp;
}

TreeVal TreeUnaryNode::compile(CodeHeap *output) {
	TreeVal x;
	int r;
	
	switch(op) {
	case OP_NEG:
		x = exp->compile(output);
		output->emit(INST_NEGINT, x.r, x.r);
		break;
	case OP_INV:
		x = exp->compile(output);
		switch (x.t) {
		case TYPE_INTEGER:	output->emit(INST_INVINT, x.r, x.r); break;
		case TYPE_PIXEL32:	output->emit(INST_INVPIXEL32, x.r, x.r); break;
		case TYPE_PIXEL64:	output->emit(INST_INVPIXEL64, x.r, x.r); break;
		}
		break;
	case OP_POSTINC:
		x = exp->compileRegValue(output);
		if (x.r<0)
			throw ParseError(offset, "only variables can be postincremented");

		if (x.t != TYPE_INTEGER)
			throw ParseError(offset, "only integers can be postincremented");

		output->emit(INST_MOVEINT, r = output->push(), x.r);
		output->emit(INST_INCINT, x.r);
		x.r = r;
		break;
	case OP_PREINC:
		x = exp->compileRegValue(output);
		if (x.r<0)
			throw ParseError(offset, "only variables can be preincremented");

		if (x.t != TYPE_INTEGER)
			throw ParseError(offset, "only integers can be preincremented");

		output->emit(INST_INCINT, x.r);
		output->emit(INST_MOVEINT, r = output->push(), x.r);
		x.r = r;
		break;
	case OP_POSTDEC:
		x = exp->compileRegValue(output);
		if (x.r<0)
			throw ParseError(offset, "only variables can be postdecremented");

		if (x.t != TYPE_INTEGER)
			throw ParseError(offset, "only integers can be postdecremented");

		output->emit(INST_MOVEINT, r = output->push(), x.r);
		output->emit(INST_DECINT, x.r);
		x.r = r;
		break;
	case OP_PREDEC:
		x = exp->compileRegValue(output);
		if (x.r<0)
			throw ParseError(offset, "only variables can be predecremented");

		if (x.t != TYPE_INTEGER)
			throw ParseError(offset, "only integers can be predecremented");

		output->emit(INST_DECINT, x.r);
		output->emit(INST_MOVEINT, r = output->push(), x.r);
		x.r = r;
		break;
	case OP_RED:
	case OP_GREEN:
	case OP_BLUE:
		x = exp->compile(output);
		if (x.t == TYPE_PIXEL32) {
			switch(op) {
			case OP_RED:	output->emit(INST_GETRPIXEL32, x.r, x.r); break;
			case OP_GREEN:	output->emit(INST_GETGPIXEL32, x.r, x.r); break;
			case OP_BLUE:	output->emit(INST_GETBPIXEL32, x.r, x.r); break;
			}
		} else if (x.t == TYPE_PIXEL64) {
			switch(op) {
			case OP_RED:	output->emit(INST_GETRPIXEL64, x.r, x.r); break;
			case OP_GREEN:	output->emit(INST_GETGPIXEL64, x.r, x.r); break;
			case OP_BLUE:	output->emit(INST_GETBPIXEL64, x.r, x.r); break;
			}
			output->pop();
		} else
			throw ParseError(offset, "channel extraction operators only valid for pixel32 and pixel64 types");

		x.t = TYPE_INTEGER;

		break;

	case OP_CASTPIXEL32:
		x = exp->compile(output);
		if (x.t == TYPE_PIXEL64) {
			output->emit(INST_PIXEL64TOPIXEL32, x.r, x.r);
			output->pop();
		} else if (x.t != TYPE_PIXEL32)
			throw ParseError(offset, "cannot convert expression to type 'pixel32'");

		x.t = TYPE_PIXEL32;
		break;

	case OP_CASTPIXEL64:
		x = exp->compile(output);
		if (x.t == TYPE_PIXEL32) {
			output->emit(INST_PIXEL32TOPIXEL64, x.r, x.r);
			output->push();
		} else if (x.t != TYPE_PIXEL64)
			throw ParseError(offset, "cannot convert expression to type 'pixel64'");

		x.t = TYPE_PIXEL64;
		break;
	}

	return x;
}

TreeVal TreeBinaryNode::compile(CodeHeap *output) {
	TreeVal x = left->compile(output);
	TreeVal y = right->compile(output);

	// type check

	switch(op) {
	case OP_ASSIGN:
		break;

	case OP_ADD:	// available for pixel32/pixel64/int
	case OP_SUB:	// promote to widest
	case OP_MUL:
	case OP_AND:
	case OP_OR:
	case OP_XOR:

		if (x.t == TYPE_BITMAP || y.t == TYPE_BITMAP) {
			throw ParseError(offset, "cannot apply operator to type 'bitmap'");
		} else if (x.t == TYPE_VOID || y.t == TYPE_VOID) {
			throw ParseError(offset, "cannot apply operator to type 'void'");
		} else if (x.t == TYPE_PIXEL64 || y.t == TYPE_PIXEL64) {
			int rx = x.r;
			int ry = y.r;
			int popcnt = 0;
			int pushcnt = 0;

			if (x.t == TYPE_PIXEL32) {
				rx = output->push(2);
				popcnt += 2;
				pushcnt = 1;
				output->emit(INST_PIXEL32TOPIXEL64, rx, x.r);
			} else if (x.t != TYPE_PIXEL64)
				throw ParseError(offset, "cannot convert expression to type 'pixel64'");

			if (y.t == TYPE_PIXEL32) {
				ry = output->push(2);
				popcnt += 2;
				output->emit(INST_PIXEL32TOPIXEL64, ry, y.r);
			} else if (y.t != TYPE_PIXEL64)
				throw ParseError(offset, "cannot convert expression to type 'pixel64'");

			x.t = TYPE_PIXEL64;

			switch(op) {
			case OP_ADD: output->emit(INST_ADDPIXEL64, x.r, rx, ry); break;
			case OP_SUB: output->emit(INST_SUBPIXEL64, x.r, rx, ry); break;
			case OP_MUL: output->emit(INST_MULPIXEL64, x.r, rx, ry); break;
			case OP_AND: output->emit(INST_ANDPIXEL64, x.r, rx, ry); break;
			case OP_OR:  output->emit(INST_ORPIXEL64, x.r, rx, ry); break;
			case OP_XOR: output->emit(INST_XORPIXEL64, x.r, rx, ry); break;
			}

			if (popcnt)
				output->pop(popcnt);

			if (pushcnt)
				output->push();

		} else if (x.t == TYPE_PIXEL32 || y.t == TYPE_PIXEL32) {
			if (x.t != TYPE_PIXEL32)
				throw ParseError(offset, "cannot convert expression to type 'pixel32'");

			if (y.t != TYPE_PIXEL32)
				throw ParseError(offset, "cannot convert expression to type 'pixel32'");

			switch(op) {
			case OP_ADD: output->emit(INST_ADDPIXEL32, x.r, x.r, y.r); break;
			case OP_SUB: output->emit(INST_SUBPIXEL32, x.r, x.r, y.r); break;
			case OP_MUL: output->emit(INST_MULPIXEL32, x.r, x.r, y.r); break;
			case OP_AND: output->emit(INST_ANDPIXEL32, x.r, x.r, y.r); break;
			case OP_OR:  output->emit(INST_ORPIXEL32, x.r, x.r, y.r); break;
			case OP_XOR: output->emit(INST_XORPIXEL32, x.r, x.r, y.r); break;
			}

			x.t = TYPE_PIXEL32;

		} else {
			switch(op) {
			case OP_ADD: output->emit(INST_ADDINT, x.r, x.r, y.r); break;
			case OP_SUB: output->emit(INST_SUBINT, x.r, x.r, y.r); break;
			case OP_MUL: output->emit(INST_MULINT, x.r, x.r, y.r); break;
			case OP_AND: output->emit(INST_ANDINT, x.r, x.r, y.r); break;
			case OP_OR:  output->emit(INST_ORINT, x.r, x.r, y.r); break;
			case OP_XOR: output->emit(INST_XORINT, x.r, x.r, y.r); break;
			}
		}

		break;

	case OP_DIV:	// available for integer only
	case OP_CMPLT:
	case OP_CMPGT:
	case OP_CMPLE:
	case OP_CMPGE:
	case OP_CMPEQ:
	case OP_CMPNE:
	case OP_LAND:
	case OP_LOR:
		if (x.t != TYPE_INTEGER)
			throw ParseError(offset, "expression is not of type int");

		if (y.t != TYPE_INTEGER)
			throw ParseError(offset, "expression is not of type int");

		switch(op) {
		case OP_DIV:	output->emit(INST_DIVINT, x.r, x.r, y.r);	break;
		case OP_CMPLT:	output->emit(INST_CMPLTINT, x.r, x.r, y.r);	break;
		case OP_CMPGT:	output->emit(INST_CMPGTINT, x.r, x.r, y.r);	break;
		case OP_CMPLE:	output->emit(INST_CMPLEINT, x.r, x.r, y.r);	break;
		case OP_CMPGE:	output->emit(INST_CMPGEINT, x.r, x.r, y.r);	break;
		case OP_CMPEQ:	output->emit(INST_CMPEQINT, x.r, x.r, y.r);	break;
		case OP_CMPNE:	output->emit(INST_CMPNEINT, x.r, x.r, y.r);	break;
		case OP_LAND:	output->emit(INST_LANDINT, x.r, x.r, y.r);	break;
		case OP_LOR:	output->emit(INST_LORINT, x.r, x.r, y.r);	break;
		}

		break;

	case OP_LSHIFT:
	case OP_RSHIFT:
	case OP_RLSHIFT:
		if (y.t != TYPE_INTEGER)
			throw ParseError(offset, "shift count must be an integer");

		if (x.t == TYPE_INTEGER) {
			switch(op) {
			case OP_LSHIFT:		output->emit(INST_LSHIFTINT, x.r, x.r, y.r); break;
			case OP_RSHIFT:		output->emit(INST_RSHIFTINT, x.r, x.r, y.r); break;
			case OP_RLSHIFT:	output->emit(INST_RLSHIFTINT, x.r, x.r, y.r); break;
			}
		} else if (x.t == TYPE_PIXEL64) {
			switch(op) {
			case OP_LSHIFT:		output->emit(INST_LSHIFTPIXEL64, x.r, x.r, y.r); break;
			case OP_RSHIFT:		output->emit(INST_RSHIFTPIXEL64, x.r, x.r, y.r); break;
			case OP_RLSHIFT:	output->emit(INST_RLSHIFTPIXEL64, x.r, x.r, y.r); break;
			}
		} else
			throw ParseError(offset, "only types 'pixel64' and 'int' can be shifted");

		break;
	}

	output->pop();
	if (y.t == TYPE_PIXEL64)
		output->pop();

	return x;
}

TreeVal TreeTernaryNode::compile(CodeHeap *output) {
	TreeVal x = cond->compile(output), y;
	int patchip1, patchip2;

	if (x.t != TYPE_INTEGER)
		throw ParseError(offset, "condition must be of type integer");

	output->pop();
	patchip1 = output->emit(INST_JZ, 0, x.r);

	x = truepath->compile(output);

	patchip2 = output->emit(INST_JMP);

	(*output)[patchip1].r2 = output->getIP();

	y = falsepath->compile(output);

	(*output)[patchip2].r2 = output->getIP();

	if (x.t != y.t)
		throw ParseError(offset, "last two expressions must be the same type in a ?:");

	return x;
}

TreeVal TreeAssignNode::compile(CodeHeap *output) {
	TreeVal v = right->compile(output);

	left->compileLValue(output, v);

	return v;
}

///////////////////////////////////////////////////////////////////////////
//
//	basic code generation -- statements
//
///////////////////////////////////////////////////////////////////////////

void TreeExpStatementNode::compile(CodeHeap *output) {
	TreeVal x = exp->compile(output);

	output->pop();
	if (x.t == TYPE_PIXEL64)
		output->pop();

	if (next)
		next->compile(output);
}

void TreeIfNode::compile(CodeHeap *output) {
	TreeVal x = cond->compile(output);
	int patchip1, patchip2;

	if (x.t != TYPE_INTEGER)
		throw ParseError(offset, "condition must be of type integer");

	output->pop();

	patchip1 = output->emit(INST_JZ, 0, x.r);

	truepath->compile(output);

	if (falsepath) {
		patchip2 = output->emit(INST_JMP);
		(*output)[patchip1].r2 = output->getIP();

		falsepath->compile(output);

		(*output)[patchip2].r2 = output->getIP();
	} else {
		(*output)[patchip1].r2 = output->getIP();
	}

	if (next)
		next->compile(output);
}

void TreeWhileNode::compile(CodeHeap *output) {
	TreeVal x;
	int patchip1, patchip2;

	patchip1 = output->getIP();

	x = cond->compile(output);

	if (x.t != TYPE_INTEGER)
		throw ParseError(offset, "condition must be of type integer");

	output->pop();

	patchip2 = output->emit(INST_JZ, 0, x.r);

	path->compile(output);

	output->emit(INST_JMP, 0, 0, patchip1);

	(*output)[patchip2].r2 = output->getIP();

	if (next)
		next->compile(output);
}

void TreeDoWhileNode::compile(CodeHeap *output) {
	TreeVal x;
	int patchip;

	patchip = output->getIP();

	path->compile(output);

	x = cond->compile(output);
	if (x.t != TYPE_INTEGER)
		throw ParseError(offset, "condition must be of type integer");
	output->pop();

	output->emit(INST_JZ, 0, x.r, patchip);

	if (next)
		next->compile(output);
}

void TreeReturnNode::compile(CodeHeap *output) {
	TreeFunctionNode *func = output->getFunction();
	TreeVal x;

	if (val) {
		x = val->compile(output);

		if (x.t != func->rettype)
			throw ParseError(offset, "type does not match function return type");
	}

	switch(func->rettype) {
	case TYPE_VOID:		output->emit(INST_RET, 0, 0, func->formalsize);				break;
	case TYPE_BITMAP:	output->emit(INST_RETBITMAP, 0, x.r, func->formalsize);		break;
	case TYPE_INTEGER:	output->emit(INST_RETINT, 0, x.r, func->formalsize);		break;
	case TYPE_PIXEL32:	output->emit(INST_RETPIXEL32, 0, x.r, func->formalsize);	break;
	case TYPE_PIXEL64:	output->emit(INST_RETPIXEL64, 0, x.r, func->formalsize);	break;
	}
}

void TreeScanNode::compile(CodeHeap *output) {
	int rinout = output->push(bitmap2exp ? 8 : 6);
	TreeVal rbmp1 = bitmap1exp->compile(output);
	TreeVal rbmp2;
	int patchip;

	activeidx = rinout;

	if (!bitmap2exp) {
		patchip = output->emit(INST_SCANSTART1, rinout, rbmp1.r);
	} else {
		rbmp2 = bitmap2exp->compile(output);

		patchip = output->emit(INST_SCANSTART2, rinout, rbmp1.r, rbmp2.r);
	}

	if (code)
		code->compile(output);

	if (bitmap2exp) {
		output->emit(INST_SCANEND2, 0, rinout, patchip+1);
	} else {
		output->emit(INST_SCANEND1, 0, rinout, patchip+1);
	}

	output->pop(bitmap2exp?8:6);
}

void TreeFunctionNode::compile(CodeHeap *output) {
	int patchip;

	ip = patchip = output->emit(INST_LINK, 0, 0, localsize);

	output->setFunction(this);
	output->setSP(localsize);

	if (code)
		code->compile(output);

	output->emit(INST_RET, 0, 0, formalsize);

	(*output)[patchip].r1 = output->getSPLimit();
}

void TreeTranslationNode::compile(CodeHeap *output) {
	func->compile(output);
}

Instruction *CodeGenerator::Compile() {
	TreeTranslationNode *pNode;
	CodeHeap ch;
	int i;

	ch.setFunctionList(pUnit);

	for(pNode = pUnit; pNode; pNode = pNode->next)
		pNode->compile(&ch);

	// backpatch all calls

	for(i=0; i<ch.getIP(); ++i) {
		Instruction& inst = ch[i];

		if ((int)inst.opcode >= (int)INST_CALL && (int)inst.opcode < (int)INST_RET)
			inst.r2 = ((TreeFunctionNode *)inst.r2)->ip;
	}

	// dump code

	extern void Disassemble(int ip, const Instruction&);

	for(i=0; i<ch.getIP(); ++i) {
		Disassemble(i, ch[i]);
	}

	return ch.getCode();
}
