//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_FILTFACTORY_COMPILE_H
#define f_FILTFACTORY_COMPILE_H

#include <string.h>
#include <stdlib.h>

#include "parser.h"

///////////////////////////////////////////////////////////////////////////

class TreeFunctionVariableNode;
class TreeScanNode;
class CodeHeap;

///////////////////////////////////////////////////////////////////////////

struct TreeVal {
	int r;
	Type t;

	TreeVal() : t(TYPE_VOID) {}
	TreeVal(int _r, Type _t) : r(_r), t(_t) {}
};

class TreeValueNode : public ParseNode {
public:
	TreeValueNode(ParseNodeAllocator *a, int offset) : ParseNode(a, offset) {}

	virtual TreeVal compile(CodeHeap *)=0;
};

class TreeVoidNode : public ParseNode {
public:
	TreeVoidNode(ParseNodeAllocator *a, int offset) : ParseNode(a, offset) {}

	virtual void compile(CodeHeap *)=0;
};

class TreeExpressionNode : public TreeValueNode {
public:
	TreeExpressionNode(ParseNodeAllocator *a, int offset) : TreeValueNode(a, offset) {}

	virtual TreeVal compileLValue(CodeHeap *, TreeVal v)=0;
	virtual TreeVal compileRegValue(CodeHeap *output);
};

class TreeConstantNode : public TreeExpressionNode {
public:
	union {
		unsigned long pix;
		int i;
	} val;

	Type type;

	TreeConstantNode(ParseNodeAllocator *a, int offset, unsigned long pix) : TreeExpressionNode(a, offset), type(TYPE_PIXEL32) {
		val.pix = pix;
	}
	TreeConstantNode(ParseNodeAllocator *a, int offset, int i) : TreeExpressionNode(a, offset), type(TYPE_INTEGER) {
		val.i = i;
	}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeVRefNode : public TreeExpressionNode {
public:
	TreeFunctionVariableNode *varnode;

	TreeVRefNode(ParseNodeAllocator *a, int offset, TreeFunctionVariableNode *_varnode) : TreeExpressionNode(a, offset), varnode(_varnode) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
	TreeVal compileRegValue(CodeHeap *output);
};

class TreeInNode : public TreeExpressionNode {
public:
	TreeScanNode *scanner;

	TreeInNode(ParseNodeAllocator *a, int offset, TreeScanNode *_scanner) : TreeExpressionNode(a, offset), scanner(_scanner) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeOutNode : public TreeExpressionNode {
public:
	TreeScanNode *scanner;

	TreeOutNode(ParseNodeAllocator *a, int offset, TreeScanNode *_scanner) : TreeExpressionNode(a, offset), scanner(_scanner) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
	TreeVal compileRegValue(CodeHeap *output);
};

class TreeCallArgumentNode : public ParseNode {
public:
	TreeCallArgumentNode *nextarg;
	TreeExpressionNode *exp;

	TreeCallArgumentNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_expnode) : ParseNode(a, offset), exp(_expnode) {}

	TreeVal compile(CodeHeap *output);
};

class TreeFunctionCallNode : public TreeExpressionNode {
public:
	TreeCallArgumentNode *args;
	char *name;

	TreeFunctionCallNode(ParseNodeAllocator *a, int offset, char *_name) : TreeExpressionNode(a, offset), args(NULL), name(strdup(_name)) {}
	~TreeFunctionCallNode() {
		free(name);
	}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeIndexingNode : public TreeExpressionNode {
public:
	TreeExpressionNode *exp;
	TreeExpressionNode *xexp, *yexp;

	TreeIndexingNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_exp, TreeExpressionNode *_xexp,  TreeExpressionNode *_yexp)
		: TreeExpressionNode(a, offset), exp(_exp), xexp(_xexp), yexp(_yexp) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeUnaryNode : public TreeExpressionNode {
public:
	TreeExpressionNode *exp;
	Operator op;

	TreeUnaryNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_exp, Operator _op)
		: TreeExpressionNode(a, offset), exp(_exp), op(_op) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeBinaryNode : public TreeExpressionNode {
public:
	TreeExpressionNode *left, *right;
	Operator op;

	TreeBinaryNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_left, TreeExpressionNode *_right,
		Operator _op) : TreeExpressionNode(a, offset), left(_left), right(_right), op(_op) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeTernaryNode : public TreeExpressionNode {
public:
	TreeExpressionNode *cond, *truepath, *falsepath;

	TreeTernaryNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_cond, TreeExpressionNode *_truepath, TreeExpressionNode *_falsepath)
		: TreeExpressionNode(a, offset), cond(_cond), truepath(_truepath), falsepath(_falsepath) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

class TreeAssignNode : public TreeExpressionNode {
public:
	TreeExpressionNode *left, *right;

	TreeAssignNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_left, TreeExpressionNode *_right) : TreeExpressionNode(a, offset), left(_left), right(_right) {}

	TreeVal compile(CodeHeap *output);
	TreeVal compileLValue(CodeHeap *output, TreeVal v);
};

/////// statements

class TreeStatementNode : public TreeVoidNode {
public:
	TreeStatementNode *next;

	TreeStatementNode(ParseNodeAllocator *a, int offset) : TreeVoidNode(a, offset), next(0) {}
	TreeStatementNode(ParseNodeAllocator *a, int offset, TreeStatementNode *_next) : TreeVoidNode(a, offset), next(_next) {}

};

class TreeExpStatementNode : public TreeStatementNode {
public:
	TreeExpressionNode *exp;

	TreeExpStatementNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_exp)
			: TreeStatementNode(a, offset), exp(_exp) {}

	void compile(CodeHeap *output);
};

class TreeIfNode : public TreeStatementNode {
public:
	TreeExpressionNode *cond;
	TreeStatementNode *truepath, *falsepath;

	TreeIfNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_cond,
			TreeStatementNode *_truepath, TreeStatementNode *_falsepath)
			: TreeStatementNode(a, offset), cond(_cond), truepath(_truepath), falsepath(_falsepath) {}

	void compile(CodeHeap *output);
};

class TreeWhileNode : public TreeStatementNode {
public:
	TreeExpressionNode *cond;
	TreeStatementNode *path;

	TreeWhileNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_cond, TreeStatementNode *_path)
			: TreeStatementNode(a, offset), cond(_cond), path(_path) {}

	void compile(CodeHeap *output);
};

class TreeDoWhileNode : public TreeStatementNode {
public:
	TreeStatementNode *path;
	TreeExpressionNode *cond;

	TreeDoWhileNode(ParseNodeAllocator *a, int offset, TreeStatementNode *_path, TreeExpressionNode *_cond)
		: TreeStatementNode(a, offset), path(_path), cond(_cond) {}

	void compile(CodeHeap *output);
};

class TreeReturnNode : public TreeStatementNode {
public:
	TreeExpressionNode *val;

	TreeReturnNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_val) : TreeStatementNode(a, offset), val(_val) {}

	void compile(CodeHeap *output);
};

class TreeScanNode : public TreeStatementNode {
public:
	TreeExpressionNode *bitmap1exp, *bitmap2exp;
	TreeStatementNode *code;
	int activeidx;

	TreeScanNode(ParseNodeAllocator *a, int offset, TreeExpressionNode *_b1exp, TreeExpressionNode *_b2exp)
		: TreeStatementNode(a, offset), bitmap1exp(_b1exp), bitmap2exp(_b2exp), code(0) {}

	void compile(CodeHeap *output);
};

class TreeFunctionVariableNode : public ParseNode {
public:
	TreeFunctionVariableNode *next;
	int idx;
	char *name;
	Type t;

	TreeFunctionVariableNode(ParseNodeAllocator *a, int offset, char *_name, Type _t, int _idx) : ParseNode(a, offset), name(strdup(_name)), t(_t), idx(_idx) {}
	~TreeFunctionVariableNode() {
		free(name);
	}
};

class TreeFunctionNode : public TreeVoidNode {
public:
	char *name;
	Type rettype;
	int localsize;
	int formalsize;
	int ip;

	TreeStatementNode *code;
	TreeFunctionVariableNode *formals;
	TreeFunctionVariableNode *locals;

	TreeFunctionNode(ParseNodeAllocator *a, int offset, char *_name, Type _rettype, TreeStatementNode *_code)
		: TreeVoidNode(a, offset), name(strdup(_name)), rettype(_rettype), code(_code), locals(NULL), formals(NULL)
		, localsize(0), formalsize(0) {}
	~TreeFunctionNode() {
		free(name);
	}

	void compile(CodeHeap *output);
};

class TreeTranslationNode : public TreeVoidNode {
public:
	TreeTranslationNode *next;
	TreeFunctionNode *func;

	TreeTranslationNode(ParseNodeAllocator *a, int offset, TreeFunctionNode *_func) : TreeVoidNode(a, offset), func(_func), next(0) {}

	void compile(CodeHeap *output);
};

#endif
