#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <commctrl.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"
/*
*******************************************************************************

	Virtualdub filter to radial correct the luminance
	-------------------------------------------------


		Copyright (C) 2002-2004 Alessandro Malanca, 

		http://web.tiscali.it/minomala/
 
	--------------------------------------------------

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

*******************************************************************************
*/

int radiaLumRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int radiaLumStartProc(FilterActivation *fa, const FilterFunctions *ff);
int radiaLumEndProc(FilterActivation *fa, const FilterFunctions *ff);
long radiaLumParamProc(FilterActivation *fa, const FilterFunctions *ff);
int radiaLumConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void radiaLumStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void radiaLumScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool radiaLumFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

int radiaLumInitProc(FilterActivation *fa, const FilterFunctions *ff);


///////////////////////////////////////////////////////////////////////////


#define RED_SCALE 0.299
#define GRE_SCALE 0.587
#define BLU_SCALE 0.114


typedef struct FilterParameters {

	bool	interlaced;
	int		y_correct;
	int		brightness;
	int		contrast;
	int		greenRed;
	int		yellowBlue;
	int		titleThreshold;
	int		radius;

} FilterParameters;

const FilterParameters 
	  nullPreviewParameters = {0,0,0,0,false /*,false*/ };

class FrameHistogram {

#define NUM_SAMPLES 80

public:

FrameHistogram() {
	for (int i=0; i<NUM_SAMPLES; i++) histogram[i]=1;
}

void setScaleSliderValue(int v) {
	sliderScale = 5-v;
	sliderPosition = v;
}

int getScaleSliderValue() {
	return sliderPosition;
}

void  paintHisto (HWND hdlg) {

	RECT	rHistoCtrl;

	HDC hdc;
	PAINTSTRUCT ps;

	HPEN whitePen = (HPEN)GetStockObject(WHITE_PEN);
	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);

	hdc = BeginPaint(GetDlgItem(hdlg,IDC_HISTO), &ps);

	GetClientRect(GetDlgItem(hdlg, IDC_HISTO), &rHistoCtrl);
	paintBackGr(hdc,rHistoCtrl);

	RECT rHisto;
	int lineSize = calcPaintArea (rHisto, rHistoCtrl );
	FillRect(hdc, &rHisto, (HBRUSH)GetStockObject(LTGRAY_BRUSH));

	long media;
	HPEN pen;
	pen = CreatePen( PS_SOLID,1,RGB(13,176,255));

	if ( sliderScale!=0) {
		media = rHisto.bottom - (rHisto.bottom+rHisto.top)/2;
	} else {
		media =  scaleValue((double)mediaLum,rHisto.bottom-rHisto.top);
	}


	int xPos = rHisto.left;
	for (int j=0;j<NUM_SAMPLES;j++) {
		int h = scaleValue((double)histogram[j],rHisto.bottom-rHisto.top);
		for ( int l=0; l<lineSize; l++) {
			SelectObject( hdc, pen);
			MoveToEx(hdc, xPos+l, rHisto.bottom, NULL);
			if (h<media) {			
				LineTo(hdc,	xPos+l, rHisto.bottom-h);
				SelectObject( hdc, blackPen);
				LineTo(hdc,	xPos+l, rHisto.bottom-media);
			} else {
				LineTo(hdc,	xPos+l, rHisto.bottom-media);
				SelectObject( hdc, whitePen);
				LineTo(hdc,	xPos+l, rHisto.bottom-h);
			}
		}
		xPos += lineSize;
	}

	// draw media line

	{
		HPEN rPen = CreatePen( PS_SOLID,2,RGB(255,50,70));
		SelectObject( hdc, rPen);
		MoveToEx(hdc, rHisto.left, rHisto.bottom-media, NULL);
		LineTo(hdc,	rHisto.right, rHisto.bottom-media);
		DeleteObject(rPen); 
	}

	DeleteObject(pen); 
	EndPaint(hdlg, &ps);
	return;
}


void invalidate(HWND hdlg)
{
	InvalidateRgn(hdlg,NULL,FALSE);
}

void calculateHisto(const FilterActivation *fa) {

	int cw = fa->dst.w/2;
	int ch = fa->dst.h/2;
	double max = sqrt(cw*cw+ch*ch);

	max *= 0.95; // some more pixels in boder ...

	for ( int i=0; i<NUM_SAMPLES; i++ ) totLum[i]=numVal[i]=0;

	Pixel32 *dst =  (Pixel32 *)fa->dst.data;

    int y = fa->dst.h;
	do {int x = fa->dst.w;
		do {
			Pixel32 pixel = *dst++;

			int r = (pixel & 0xFF0000)>>16;
			int g = (pixel & 0x00FF00)>>8;
			int b = (pixel & 0x0000FF);
		
			long lum = (long)((RED_SCALE * r) + (GRE_SCALE * g) + (BLU_SCALE * b)); 
			double d = sqrt((x-cw)*(x-cw)+(y-ch)*(y-ch)) +1 ; // +1 ????
			d = d/max*(double)(NUM_SAMPLES-1);

			if (d>= (double)NUM_SAMPLES) d= (double)(NUM_SAMPLES-1);
			totLum[(int) d] += lum;
			numVal[(int) d]++;

		} while(--x);
		dst = (Pixel32 *)((char *)dst + fa->dst.modulo);
    } while(--y);

	mediaLum=0;
	long numMedia=0;

	for( int j=0; j<(double)NUM_SAMPLES;j++) {
		int t = (numVal[j]>0) ? numVal[j] : 1;
		histogram[j] = totLum[j] / t;
		mediaLum += totLum[j];
		numMedia += numVal[j];
	}
	mediaLum /= numMedia;
}

private:

	long	totLum[NUM_SAMPLES];
	long	numVal[NUM_SAMPLES];
	int		histogram[NUM_SAMPLES];
	long	mediaLum;
	int		sliderScale; // 0 .. 5
	int		sliderPosition;

void paintBackGr( HDC hdc, RECT r ) {

	FillRect(hdc, &r, (HBRUSH)GetStockObject(LTGRAY_BRUSH));

	HPEN blackPen = (HPEN)GetStockObject(BLACK_PEN);
	SelectObject( hdc, blackPen);

	MoveToEx(hdc, r.left, r.top, NULL);
	LineTo(hdc,	r.right-1,r.top);
	LineTo(hdc,	r.right-1,r.bottom-1);
	LineTo(hdc,	r.left,r.bottom-1);
	LineTo(hdc,	r.left,r.top);
}

int calcPaintArea ( RECT & res, RECT rCtrl ) {
	int step = (rCtrl.right-rCtrl.left) / NUM_SAMPLES;
	int left = (rCtrl.right - rCtrl.left - (step * NUM_SAMPLES)) / 2;
	res.left = rCtrl.left+left;
	res.right = rCtrl.left+left+step*NUM_SAMPLES;
	res.bottom = rCtrl.bottom-4;
	res.top = rCtrl.top+4;
	return step;
}

int scaleValue(double histo, int max) {

#define FIT(v)  ( (double)v / 255.0 * (double)max )

	int res;
	res = (int)FIT(histo);
	if ( sliderScale == 0) {
		res = (int)FIT(histo);
	} else {
		double value = histo - (double)mediaLum;		
		double scale = 1.0 + ((double)(sliderScale-1) * 0.25);
		res = (int)(((double)max/2.0)+FIT(value*scale)) ;
	}
	if (res<0) res = 0;
	if (res>=max) res = max-1;
	return res;
#undef FIT
}

}; // FrameHistogram

typedef struct PixelAdjust {

	Pixel32 red[256];
	Pixel32 green[256];
	Pixel32 blue[256];

} PixelAdjust;


typedef struct MyFilterData {

	int					**luma_change;	
	bool				disableFilter;
	FilterParameters	par,save;
	IFilterPreview		*ifp;
	FrameHistogram      frameHistogram;
	bool				displayingPanel;
	PixelAdjust			*pixelAdjust; 

} MyFilterData;


ScriptFunctionDef radiaLum_func_defs[]={
    { (ScriptFunctionPtr)radiaLumScriptConfig, "Config", "0iiiiiiii" },
    { NULL },
};

CScriptObject radiaLum_obj={
    NULL, radiaLum_func_defs
};

struct FilterDefinition filterDef_radiaLum = {

    NULL, NULL, NULL,			// next, prev, module
    "radial luminance [2.0]",	// name
    "Correct radial luminance for 8mm or Super8 films.\nUses the tstudent distribution and allows rgb color correction.",
								// desc
    "Alessandro Malanca",		// maker
    NULL,						// private_data
    sizeof(MyFilterData),		// inst_data_size
    radiaLumInitProc,			// initProc
    NULL,						// deinitProc
    radiaLumRunProc,			// runProc
    NULL,						// ParamProc
    radiaLumConfigProc,			// ConfigProc
    radiaLumStringProc,			// stringProc
    radiaLumStartProc,			// startProc
    radiaLumEndProc,			// endProc
    &radiaLum_obj,				// script_obj
    radiaLumFssProc,			// fssProc

};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_radiaLum;

///////////////////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_radiaLum = ff->addFilter(fm, &filterDef_radiaLum, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_radiaLum);
}

///////////////////////////////////////////////////////////////////////////


int radiaLumInitProc(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	mfd->displayingPanel = false;
	mfd->par.titleThreshold = 200;
	return 0;
}


PixelAdjust * prepareToAdjust(FilterParameters par) {

#define P_SCALE 2

	long dr_adjust =  par.greenRed - par.yellowBlue/2;
	long dg_adjust =  - par.greenRed - par.yellowBlue/2;
	long db_adjust =  par.yellowBlue;

	if ( dr_adjust<0 ) dr_adjust /= P_SCALE;
	if ( dg_adjust<0 ) dg_adjust /= P_SCALE;
	if ( db_adjust<0 ) db_adjust /= P_SCALE;

    dr_adjust += 256;
	dg_adjust += 256;
	db_adjust += 256;


#undef P_SCALE 


//	long scale_r, scale_g, scale_b;

	long brightness	= par.brightness;
	long contrast	= par.contrast + 256;

	long r,g,b;
	
	PixelAdjust * res;
	res = new PixelAdjust();
	for (int i=0; i<256; i++) {

		r = g = b = i;

		// adjust brightness
		r += brightness;
		g += brightness;
		b += brightness;

		// adjust contrast

#define CM 120
		r = ( ( (r-CM) * contrast ) / 256) + CM ;
		g = ( ( (g-CM) * contrast ) / 256) + CM ;
		b = ( ( (b-CM) * contrast ) / 256) + CM ;
#undef CM

		// adjust colors & radial

		r = (r * dr_adjust) / 256; 
		g = (g * dg_adjust) / 256; 
		b = (b * db_adjust) / 256;

		if (r < 0) r=0; else if (r > 255) r=255;
		if (g < 0) g=0; else if (g > 255) g=255;
		if (b < 0) b=0; else if (b > 255) b=255;

		res->red[i]		= r<<16;
		res->green[i]	= g<<8;
		res->blue[i]	= b;

	}
	return res;

}


double tStudent3 (double x, double y ) {
	x *= 2; y *= 2;
	return  1.0 / ( 1.0 + ((x*x)+(y*y)) );
}


int radiaLumStartProc(FilterActivation *fa, const FilterFunctions *ff) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	int cols = fa->src.w+1;
	int rows = fa->src.h+1;

	if (mfd->par.interlaced) rows = rows/2;

	int cx = cols/2; cols = cx*2;
	int cy = rows/2; rows = cy*2;

	double radius = ( 100.0 + (double) mfd->par.radius ) / 100.0;
	double fInScale = cx * radius;

	// mfd->par.radius in -50 +50;
	double radius_correction = + 0.2 * ( mfd->par.radius/50.0); // -0.2 ..+0.2

	// ----------------------------------
	//  Allocate param tables
	// ----------------------------------	
	int  * temp;
	if (!(temp = new int[cols*rows])) return 1;

	if (mfd->par.interlaced) {
		if (!(mfd->luma_change = new int * [rows*2])) return 1;
	} else {
		if (!(mfd->luma_change = new int * [rows])) return 1;	
	}

	for ( int i =0; i<rows; i++) {
		mfd->luma_change[i]= &temp[i*cols];
	}

	int *row, *row2;
	int x,y;

	// ----------------------------------
	// Fill table	
	// ----------------------------------

	double correct = (double)mfd->par.y_correct * 1.5 ;

	for ( y=0; y<cy; y++ ) {

		row = mfd->luma_change[cy-y-1];
		row2 = mfd->luma_change[cy+y];

		for ( x=0; x<cx; x++ ) {

			double fun = 0.5 + radius_correction - tStudent3 
				(	(double)x/(double)fInScale,	(double)y/(double)fInScale) ;

			fun *= correct;
			if ( fun > 0.0 ) fun *= 3.0;
			row[cx+x]	=  row[cx-x-1]	=
			row2[cx+x]	=  row2[cx-x-1] =   (int) fun;
		}	
	}

	if (mfd->par.interlaced) {
		int k;
		for (i =0, k=0; i<rows; i++, k+=2) {
			mfd->luma_change[k] = mfd->luma_change[k+1]= &temp[i*cols];
		}
	}

	mfd->pixelAdjust = prepareToAdjust(mfd->par);

    return 0;

}

bool testIsTitle(const long soglia, Pixel * src, int width) {
	Pixel32 * p = src;
	long media=0;
	for (int i=0; i<5; i++) {
		Pixel32 pix = *p++;
		media +=	((pix & 0xFF0000)>>16)
					+ ((pix & 0x00FF00)>>8)
					+ (pix & 0x0000FF);
	}
	p += width-11;
	for ( i=0; i<5; i++) {
		Pixel32 pix = *p++;
		media +=	((pix & 0xFF0000)>>16)
					+ ((pix & 0x00FF00)>>8)
					+ (pix & 0x0000FF);
	}
	media /= 10*3;
	return media > soglia;
}


int radiaLumRunProc(const FilterActivation *fa, const FilterFunctions *ff) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    PixDim w, h;
    Pixel32 *src, *dst;

	if (mfd->disableFilter) {
		fa->dst.data = fa->src.data;
		mfd->frameHistogram.calculateHisto(fa);
		return 0;
	}

    src = (Pixel32 *)fa->src.data;
    dst = (Pixel32 *)fa->dst.data;

	long r, g, b; 

	int *rr;

	Pixel32 * red_pix = mfd->pixelAdjust->red;
	Pixel32 * gre_pix = mfd->pixelAdjust->green;
	Pixel32 * blu_pix = mfd->pixelAdjust->blue;

	Pixel32 old_pixel;

	bool isTile = testIsTitle(mfd->par.titleThreshold,src,fa->src.w);
	int scale = 0;

    h = fa->src.h;
    do { 
		w = fa->src.w;
		rr = mfd->luma_change[h-1];
		do {

			old_pixel = *src++;

			// get source pixel colors
			r = (old_pixel & 0xFF0000)>>16;
			g = (old_pixel & 0x00FF00)>>8;
			b =  old_pixel & 0x0000FF;

			scale = rr[w-1];
			
			if ((scale !=0) && (!isTile)) {

				// adjust radial luminance
				int max = r;
				if (g>max) max = g;
				if (b>max) max = b;
				if ((max+scale) >255) {
					scale = 255-max;
				}
				scale += 256;
				r = (r * scale) >> 8;
				g = (g * scale) >> 8;
				b = (b * scale) >> 8;
			}

			// adjust brightness, contrast and colors

			*dst++ = red_pix[r] | gre_pix[g] | blu_pix[b];

		} while(--w);

	    src = (Pixel32 *)((char *)src + fa->src.modulo);
		dst = (Pixel32 *)((char *)dst + fa->dst.modulo);

    } while(--h);

	if (mfd->displayingPanel) {
		mfd->frameHistogram.calculateHisto(fa);
	}
    return 0;
}


int radiaLumEndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	delete[] mfd->luma_change[0];
	delete[] mfd->luma_change;
	mfd->luma_change = NULL;

	delete mfd->pixelAdjust;
	mfd->pixelAdjust=NULL;

    return 0;
}


void setDialogFeedbackValues (HWND hdlg, MyFilterData *mfd) {
		
	SetDlgItemInt(hdlg, IDC_YCORR_VALUE, mfd->par.y_correct, TRUE);
	SetDlgItemInt(hdlg, IDC_CONTRAST_VALUE, mfd->par.contrast, TRUE);
	SetDlgItemInt(hdlg, IDC_BRIGHT_VALUE, mfd->par.brightness, TRUE);
	SetDlgItemInt(hdlg, IDC_GREENRED_VALUE, mfd->par.greenRed, TRUE);
	SetDlgItemInt(hdlg, IDC_YELLOWBLUE_VALUE, mfd->par.yellowBlue, TRUE);
	SetDlgItemInt(hdlg, IDC_RADIUS_VALUE, mfd->par.radius, TRUE);

}


void redoSystem( MyFilterData * mfd, HWND hdlg) {		

	mfd->ifp->UndoSystem();
	mfd->ifp->RedoSystem();
	setDialogFeedbackValues(hdlg,mfd);
	mfd->frameHistogram.invalidate(hdlg);
}


void redoFrame ( MyFilterData * mfd, HWND hdlg ) {

//	redoSystem( mfd, hdlg);

	mfd->ifp->RedoFrame();
	setDialogFeedbackValues(hdlg,mfd);
	mfd->frameHistogram.invalidate(hdlg);

}

void getHorSliderValues(MyFilterData *mfd,HWND hdlg) {

	int y_correct = SendMessage(GetDlgItem(hdlg, IDC_YCORRECT_SLIDER), TBM_GETPOS, 0, 0);
	if ( y_correct != mfd->par.y_correct ) {
		mfd->par.y_correct = y_correct;
		redoSystem( mfd, hdlg );
		return ;
	}

	int contrast = SendMessage(GetDlgItem(hdlg, IDC_CONTRAST_SLIDER), TBM_GETPOS, 0, 0);
	if ( contrast != mfd->par.contrast ) {
		mfd->par.contrast = contrast;
		redoSystem( mfd, hdlg );
		return;
	}

	int brightness = SendMessage(GetDlgItem(hdlg, IDC_BRIGHT_SLIDER), TBM_GETPOS, 0, 0);
	if ( brightness != mfd->par.brightness ) {
		mfd->par.brightness = brightness;
		redoSystem( mfd, hdlg );
		return;
	}
		
	int greenRed = SendMessage(GetDlgItem(hdlg, IDC_GREENRED_SLIDER), TBM_GETPOS, 0, 0);
	if ( greenRed != mfd->par.greenRed ) {
		mfd->par.greenRed = greenRed;
		redoSystem( mfd, hdlg );
		return; 
	}
	
	int yellowBlue = SendMessage(GetDlgItem(hdlg, IDC_YELLOWBLUE_SLIDER), TBM_GETPOS, 0, 0);
	if ( yellowBlue != mfd->par.yellowBlue ) {
		mfd->par.yellowBlue = yellowBlue;
		redoSystem( mfd, hdlg );
		return;
	}

	int radius = SendMessage(GetDlgItem(hdlg, IDC_RADIUS_SLIDER), TBM_GETPOS, 0, 0);
	if ( radius != mfd->par.radius ) {
		mfd->par.radius = radius;
		redoSystem( mfd, hdlg );
		return ;
	}

}

void getVerSliderValues(MyFilterData *mfd,HWND hdlg) {

	int histoSlider = SendMessage(GetDlgItem(hdlg, IDC_HISTO_SLIDER), TBM_GETPOS, 0, 0);
	if ( histoSlider != mfd->frameHistogram.getScaleSliderValue() ) {
		 mfd->frameHistogram.setScaleSliderValue(histoSlider);
		redoFrame( mfd, hdlg );
		return; 
	}

}


static void levelsButtonCallback(bool fNewState, void *pvData)
{
	HWND hdlg = (HWND)pvData;
	EnableWindow(GetDlgItem(hdlg, IDSAMPLE), fNewState);
	EnableWindow(GetDlgItem(hdlg, IDC_HISTO_SLIDER), fNewState);
	EnableWindow(GetDlgItem(hdlg, IDC_DISABLE_CHECK), fNewState);
	HWND wHist = GetDlgItem(hdlg, IDC_HISTO);
	if (fNewState == TRUE) {
		ShowWindow(wHist, SW_SHOWNORMAL);
	} else {
		ShowWindow(wHist, SW_HIDE);
	}

}

void setPanelDefaults(HWND hdlg, MyFilterData *mfd) {

	HWND hWnd = GetDlgItem(hdlg,IDC_YCORRECT_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(0, 100));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.y_correct);

	hWnd = GetDlgItem(hdlg,IDC_RADIUS_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-50, 50));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.radius);

	hWnd = GetDlgItem(hdlg,IDC_CONTRAST_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.contrast );

	hWnd = GetDlgItem(hdlg,IDC_BRIGHT_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-50, 50));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.brightness);

	hWnd = GetDlgItem(hdlg,IDC_GREENRED_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.greenRed);

	hWnd = GetDlgItem(hdlg,IDC_YELLOWBLUE_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, mfd->par.yellowBlue);

	hWnd = GetDlgItem(hdlg,IDC_HISTO_SLIDER);
	SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(0, 5));
	SendMessage(hWnd, TBM_SETPOS, (WPARAM)TRUE, 7);

	CheckDlgButton(hdlg, IDC_INTERLACED_CHECK, mfd->par.interlaced?BST_CHECKED:BST_UNCHECKED);

	SetDlgItemInt(hdlg, IDC_TITLE_SAMPLE, mfd->par.titleThreshold, FALSE);

}



BOOL CALLBACK radiaLumConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) {

        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

			mfd->displayingPanel = true;
			EnableWindow(GetDlgItem(hdlg, IDSAMPLE), false);

			levelsButtonCallback(false, (void *)hdlg);
			mfd->ifp->SetButtonCallback(levelsButtonCallback, (void *)hdlg);

			// save for cancel ...
			mfd->save = mfd->par;
			HWND hWnd;

			hWnd = GetDlgItem(hdlg,IDPREVIEW);
			mfd->ifp->InitButton(hWnd);
			setPanelDefaults(hdlg,mfd);
			// update vales in boxes
			setDialogFeedbackValues(hdlg,mfd);

			return TRUE;


        case WM_COMMAND:

            switch(LOWORD(wParam)) {

            case IDOK:

				mfd->disableFilter = false;
				mfd->displayingPanel = false;
                EndDialog(hdlg, 0);
                return TRUE;

			case IDHELP:	{
				char prog[256];
				char path[256];
				LPTSTR ptr;

				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\RadiaLum.htm");
				ShellExecute(hdlg, "open", path, NULL, NULL, SW_SHOWNORMAL);
				return TRUE;
				}

            case IDCANCEL:
				mfd->par = mfd->save;
				mfd->disableFilter = false;
				mfd->displayingPanel = false;
                EndDialog(hdlg, 1);
                return FALSE;

			case IDPREVIEW: {
				HWND hWnd = GetDlgItem(hdlg,IDPREVIEW);
				mfd->ifp->RedoSystem();
				mfd->ifp->Toggle(hWnd);
				break;
			}
			
			case IDSAMPLE:
				mfd->frameHistogram.invalidate(hdlg);
				break;

			case IDC_COLOR_DEFAULT :
				mfd->par.greenRed = mfd->par.yellowBlue = 0;
				SendMessage(GetDlgItem(hdlg,IDC_GREENRED_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, mfd->par.greenRed);
				SendMessage(GetDlgItem(hdlg,IDC_YELLOWBLUE_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, mfd->par.yellowBlue);
				redoSystem( mfd, hdlg );
				break;

			case IDC_LUMACORR_DEFAULT:
				mfd->par.contrast = mfd->par.brightness = 0;
				SendMessage(GetDlgItem(hdlg,IDC_CONTRAST_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, mfd->par.contrast );
				SendMessage(GetDlgItem(hdlg,IDC_BRIGHT_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, mfd->par.brightness);
				redoSystem( mfd, hdlg );
				break;

			case IDC_RADIALCORR_DEFAULT:
				mfd->par.radius = 0;
				SendMessage(GetDlgItem(hdlg,IDC_RADIUS_SLIDER), 
					TBM_SETPOS, (WPARAM)TRUE, mfd->par.radius);
				redoSystem( mfd, hdlg );
				break;

			case IDC_DISABLE_CHECK:
				mfd->disableFilter = ! mfd->disableFilter;
				mfd->ifp->RedoFrame();
				mfd->frameHistogram.invalidate(hdlg);
				break;

			case IDC_INTERLACED_CHECK:
				mfd->par.interlaced = ! mfd->par.interlaced;
				mfd->ifp->UndoSystem();
				mfd->ifp->RedoSystem();
				//setDialogFeedbackValues(hdlg,mfd);
				mfd->frameHistogram.invalidate(hdlg);
				break;

			case IDC_TITLE_SAMPLE: {
				int thresold = GetDlgItemInt(hdlg, IDC_TITLE_SAMPLE, NULL, FALSE);
				if (thresold != mfd->par.titleThreshold) {
					if ( thresold > 255 )  {
						thresold=255;
						SetDlgItemInt(hdlg, IDC_TITLE_SAMPLE, thresold, FALSE);
					}
					mfd->par.titleThreshold=thresold;
					mfd->ifp->RedoFrame();
					mfd->frameHistogram.invalidate(hdlg);
				}
				} break;

		}	// case WM_COMMAND:

		case WM_HSCROLL: 
			if (LOWORD(wParam) == TB_ENDTRACK)  
				getHorSliderValues(mfd,hdlg);
			break;	
			

		case WM_VSCROLL: 
			if (LOWORD(wParam) == TB_ENDTRACK)
				getVerSliderValues(mfd,hdlg);
			break;	


		case WM_PAINT:
			mfd->frameHistogram.paintHisto(hdlg);
			break;
    }

    return FALSE;
}

int radiaLumConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {

	MyFilterData * mfd = (MyFilterData *)fa->filter_data;
	mfd->ifp = fa->ifp;
	mfd->displayingPanel = false;
    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_LUM8MM), hwnd,
            radiaLumConfigDlgProc, (LPARAM)fa->filter_data
	);
}


void radiaLumStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {

    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	sprintf(str,"(cor=%d,rad=%d,inter=%s,bri=%d,con=%d,g-r=%d,y-b=%d,tit=%d )",
		mfd->par.y_correct, mfd->par.radius,
		mfd->par.interlaced ? "y": "n",
		mfd->par.brightness,  mfd->par.contrast,
		mfd->par.greenRed, mfd->par.yellowBlue,
		mfd->par.titleThreshold
	);

}

void radiaLumScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {

    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->par.y_correct = argv[0].asInt();
	mfd->par.radius = argv[1].asInt();
	mfd->par.interlaced = ((argv[2].asInt()==0) ? false : true);
	mfd->par.brightness= argv[3].asInt();
	mfd->par.contrast = argv[4].asInt();	
	mfd->par.greenRed = argv[5].asInt();
	mfd->par.yellowBlue = argv[6].asInt();
	mfd->par.titleThreshold = argv[7].asInt();

}


bool radiaLumFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    _snprintf(buf, buflen, "Config(%d, %d, %d, %d, %d, %d, %d, %d)",
		mfd->par.y_correct,  mfd->par.radius,
		(mfd->par.interlaced ? 1: 0),
		mfd->par.brightness,mfd->par.contrast, 
		mfd->par.greenRed ,	mfd->par.yellowBlue,
		mfd->par.titleThreshold
	);

    return true;
}
