//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "parser.h"
#include "compile.h"

///////////////////////////////////////////////////////////////////////////
//
//	ParseError
//
///////////////////////////////////////////////////////////////////////////

ParseError::ParseError(int offs, const char *format, ...) {
	if (pError = (ErrorString *)malloc(504 + sizeof(ErrorString))) {
		va_list val;
		int len;
		ErrorString *pError2;

		pError->refcount = 1;
		pError->offset = offs;
		va_start(val, format);
		len = _vsnprintf(pError->estr, 504, format, val);
		va_end(val);

		if (len < 0) {
			len = 503;
			pError->estr[len] = 0;
		}

		pError2 = (ErrorString *)realloc(pError, ((char *)pError->estr + len + 1) - (char *)pError);

		if (pError2)
			pError = pError2;
	}
}

ParseError::ParseError(const ParseError& x) {
	pError = x.pError;

	if (pError)
		++pError->refcount;
}

ParseError::~ParseError() {
	if (pError && !--pError->refcount)
		delete pError;
}

ParseError& ParseError::operator=(const ParseError& x) {
	if (pError && !--pError->refcount)
		delete pError;

	pError = x.pError;

	if (pError)
		++pError->refcount;

	return *this;
}

///////////////////////////////////////////////////////////////////////////
//
//	helpers
//
///////////////////////////////////////////////////////////////////////////

int Parser::getPrecedence(int tok) {
	switch(tok) {
	case '*':
	case '/':
		return 10;
	case '+':
	case '-':
		return 9;
	case TOK_LSHIFT:
	case TOK_RSHIFT:
	case TOK_RLSHIFT:
		return 8;
	case '<':
	case '>':
	case TOK_LE:
	case TOK_GE:
		return 7;
	case TOK_EQ:
	case TOK_NE:
		return 6;
	case '&':
		return 5;
	case '^':
		return 4;
	case '|':
		return 3;
	case TOK_LAND:
	case TOK_LOR:
		return 1;
	default:
		return -1;
	}
};

int Parser::getSize(Type t) {
	switch(t) {
	case TYPE_PIXEL32:
	case TYPE_INTEGER:
	case TYPE_BITMAP:
		return 1;

	case TYPE_PIXEL64:
		return 2;
	default:
		__assume(false);
	}
}

const char *Parser::getTokenName(char *buf, int tok) {
	if (!tok)
		return "end of file";
	else if (tok < 256) {
		buf[0] = '\'';
		buf[1] = (char)tok;
		buf[2] = '\'';
		buf[3] = 0;

		return buf;
	} else switch(tok) {
		case TOK_LE:		return "'<='";
		case TOK_GE:		return "'<='";
		case TOK_EQ:		return "'=='";
		case TOK_NE:		return "'!='";
		case TOK_LAND:		return "'&&'";
		case TOK_LOR:		return "'||'";
		case TOK_PLUSPLUS:		return "'++'";
		case TOK_MINUSMINUS:	return "'--'";
		case TOK_LSHIFT:	return "'<<'";
		case TOK_RSHIFT:	return "'>>'";
		case TOK_RLSHIFT:	return "'>>>'";
		case TOK_ID:		return "identifier";

		case TOK_INTEGER:	return "integer";
		case TOK_PIXVAL:	return "pixel value";

		case TOK_PIXEL32:	return "'pixel32'";
		case TOK_PIXEL64:	return "'pixel64'";
		case TOK_INT:		return "'int'";
		case TOK_BITMAP:	return "'bitmap'";
		case TOK_VOID:		return "'void'";

		case TOK_IN:		return "'in'";
		case TOK_OUT:		return "'out'";

		case TOK_SCAN:		return "'scan'";
		case TOK_RETURN:	return "'return'";
		case TOK_IF:		return "'if'";
		case TOK_WHILE:		return "'while'";
		case TOK_DO:		return "'do'";
		case TOK_ELSE:		return "'else'";
		case TOK_FOR:		return "'for'";
		default:
			return "(oops)";
	}
}

///////////////////////////////////////////////////////////////////////////
//
//	lexer
//
///////////////////////////////////////////////////////////////////////////

int Parser::nexttoken() {
	if (token_parsed >= 0)
		return token_parsed;

	while(*tokptr && isspace(*tokptr))
		++tokptr;

	tokstart = tokptr;

	if (isalpha((unsigned char)*tokptr)) {
		++tokptr;
		while(*tokptr && isalnum(*tokptr))
			++tokptr;

		switch(tokptr - tokstart) {
		case 2:
			if (!memcmp(tokstart, "if", 2))
				return token_parsed = TOK_IF;
			if (!memcmp(tokstart, "in", 2))
				return token_parsed = TOK_IN;
			if (!memcmp(tokstart, "do", 2))
				return token_parsed = TOK_DO;
			break;
		case 3:
			if (!memcmp(tokstart, "int", 3))
				return token_parsed = TOK_INT;
			if (!memcmp(tokstart, "for", 3))
				return token_parsed = TOK_FOR;
			if (!memcmp(tokstart, "out", 3))
				return token_parsed = TOK_OUT;
			break;
		case 4:
			if (!memcmp(tokstart, "scan", 4))
				return token_parsed = TOK_SCAN;
			if (!memcmp(tokstart, "void", 4))
				return token_parsed = TOK_VOID;
			if (!memcmp(tokstart, "else", 4))
				return token_parsed = TOK_ELSE;
			break;
		case 5:
			if (!memcmp(tokstart, "while", 3))
				return token_parsed = TOK_WHILE;
			break;
		case 6:
			if (!memcmp(tokstart, "bitmap", 6))
				return token_parsed = TOK_BITMAP;
			if (!memcmp(tokstart, "return", 6))
				return token_parsed = TOK_RETURN;
			break;
		case 7:
			if (!memcmp(tokstart, "pixel32", 7))
				return token_parsed = TOK_PIXEL32;
			if (!memcmp(tokstart, "pixel64", 7))
				return token_parsed = TOK_PIXEL64;
			break;
		}

		if (tokptr - tokstart > 63)
			throw ParseError(getOffset(), "identifier too long");

		memcpy(namebuf, tokstart, tokptr - tokstart);
		namebuf[tokptr - tokstart] = 0;

		return token_parsed = TOK_ID;
	} else if (tokptr[0] == '<' && tokptr[1] == '=') {
		tokptr += 2;
		return token_parsed = TOK_LE;
	} else if (tokptr[0] == '<' && tokptr[1] == '<') {
		tokptr += 2;
		return token_parsed = TOK_LSHIFT;
	} else if (tokptr[0] == '>' && tokptr[1] == '=') {
		tokptr += 2;
		return token_parsed = TOK_GE;
	} else if (tokptr[0] == '>' && tokptr[1] == '>') {
		if (tokptr[2] == '>') {
			tokptr += 3;
			return token_parsed = TOK_RLSHIFT;
		} else {
			tokptr += 2;
			return token_parsed = TOK_RSHIFT;
		}
	} else if (tokptr[0] == '=' && tokptr[1] == '=') {
		tokptr += 2;
		return token_parsed = TOK_EQ;
	} else if (tokptr[0] == '!' && tokptr[1] == '=') {
		tokptr += 2;
		return token_parsed = TOK_NE;
	} else if (tokptr[0] == '&' && tokptr[1] == '&') {
		tokptr += 2;
		return token_parsed = TOK_LAND;
	} else if (tokptr[0] == '|' && tokptr[1] == '|') {
		tokptr += 2;
		return token_parsed = TOK_LOR;
	} else if (tokptr[0] == '+' && tokptr[1] == '+') {
		tokptr += 2;
		return token_parsed = TOK_PLUSPLUS;
	} else if (tokptr[0] == '-' && tokptr[1] == '-') {
		tokptr += 2;
		return token_parsed = TOK_MINUSMINUS;
	} else if (!*tokptr)
		return token_parsed = 0;
	else if (*tokptr == '#') {
		long pix = 0;

		++tokptr;
		for(int i=0; i<6; ++i) {
			char c = *tokptr++;

			if (!isxdigit(c))
				throw ParseError(getOffset(), "not enough hex digits in pixel constant");

			pix <<= 4;
			if (isdigit(c))
				pix += c - '0';
			else
				pix += toupper(c) - 'A' + 10;
		}

		if (isxdigit(*tokptr))
			throw ParseError(getOffset(), "too many hex digits in pixel constant");

		tokval.pix = pix;

		return token_parsed = TOK_PIXVAL;
	} else if (isdigit(*tokptr)) {
		if (tokptr[0] == '0' && tokptr[1] == 'x') {
			tokval.i = 0;

			if (!isxdigit(*tokptr))
				throw ParseError(getOffset(), "expected hex constant");

			while(isxdigit(*tokptr)) {
				char c = *tokptr++;

				tokval.i <<= 4;
				
				if (isdigit(c))
					tokval.i += (int)c - '0';
				else
					tokval.i += (int)toupper(c) - 'A' + 10;
			}

			return token_parsed = TOK_INTEGER;
		} else {
			tokval.i = 0;

			while(isdigit(*tokptr))
				tokval.i = (tokval.i * 10) + ((int)*tokptr++ - '0');

			return token_parsed = TOK_INTEGER;
		}
	}

	return token_parsed = (int)(unsigned char)*tokptr++;
}

void Parser::expect(int tok) {
	int tok2 = nexttoken();

	if (tok2 != tok) {
		char buf1[4], buf2[4];

		throw ParseError(getOffset(), "expected %s, found %s", getTokenName(buf1, tok), getTokenName(buf2, tok2));
	}

	token_parsed = -1;
}

void Parser::absorb() {
	token_parsed = -1;
}

///////////////////////////////////////////////////////////////////////////
//
//	recursive descent parser
//
///////////////////////////////////////////////////////////////////////////

TreeFunctionVariableNode *Parser::getVariable(const char *name) {
	TreeFunctionVariableNode *node;
	
	for(node = pFunction->formals; node; node = node->next)
		if (!strcmp(name, node->name))
			return node;

	for(node = pFunction->locals; node; node = node->next)
		if (!strcmp(name, node->name))
			return node;

	return NULL;
}

TreeExpressionNode *Parser::ParseValue() {
	TreeExpressionNode *exp;
	int pos = getOffset();
	int tok;

	tok = nexttoken();

	if (tok == '(') {
		absorb();

		switch(tok = nexttoken()) {
		case TOK_PIXEL32:
			absorb();
			expect(')');
			return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_CASTPIXEL32);
		case TOK_PIXEL64:
			absorb();
			expect(')');
			return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_CASTPIXEL64);
		}

		exp = ParseExpression();

		expect(')');

		return exp;
	} else if (tok == TOK_ID) {
		char name[64];

		absorb();

		strcpy(name, namebuf);

		tok = nexttoken();

		if (tok == '(') {
			TreeFunctionCallNode *tcall = new TreeFunctionCallNode(autofree, pos, namebuf);

			absorb();

			if (nexttoken() != ')') {
				for(;;) {
					pos = getOffset();
					TreeExpressionNode *exp = ParseExpression();
					TreeCallArgumentNode *arg = new TreeCallArgumentNode(autofree, pos, exp);

					arg->nextarg = tcall->args;
					tcall->args = arg;

					if (nexttoken() != ',')
						break;

					absorb();
				}
			}

			expect(')');

			return tcall;
		} else {
			TreeFunctionVariableNode *tfvn = getVariable(name);

			if (!tfvn)
				throw ParseError(getOffset(), "variable '%s' not declared", name);

			return new TreeVRefNode(autofree, pos, tfvn);
		}
	} else if (tok == TOK_PIXVAL) {
		absorb();
		return new TreeConstantNode(autofree, pos, tokval.pix);
	} else if (tok == TOK_INTEGER) {
		absorb();
		return new TreeConstantNode(autofree, pos, tokval.i);
	} else if (tok == TOK_IN) {
		absorb();
		if (!pScanner)
			throw ParseError(getOffset(), "'in' can only be used within scan statements");

		return new TreeInNode(autofree, pos, pScanner);
	} else if (tok == TOK_OUT) {
		absorb();
		if (!pScanner)
			throw ParseError(getOffset(), "'out' can only be used within scan statements");

		return new TreeOutNode(autofree, pos, pScanner);
	} else {
		char buf[4];

		throw ParseError(getOffset(), "expected value, found %s", getTokenName(buf, tok));
	}

}

TreeExpressionNode *Parser::ParsePostfixExpression() {
	TreeExpressionNode *exp;
	int pos = getOffset();
	
	exp = ParseValue();

	for(;;) {
		switch(nexttoken()) {
		case '[':
			{
				absorb();
				TreeExpressionNode *xexp = ParseExpression();
				expect(',');
				TreeExpressionNode *yexp = ParseExpression();
				expect(']');

				exp = new TreeIndexingNode(autofree, pos, exp, xexp, yexp);
			}
			break;

		case TOK_PLUSPLUS:
			absorb();
			exp = new TreeUnaryNode(autofree, pos, exp, OP_POSTINC);
			break;

		case TOK_MINUSMINUS:
			absorb();
			exp = new TreeUnaryNode(autofree, pos, exp, OP_POSTDEC);
			break;

		case '.':
			absorb();
			expect(TOK_ID);

			if (!strcmp(namebuf, "r"))
				exp = new TreeUnaryNode(autofree, pos, exp, OP_RED);
			else if (!strcmp(namebuf, "g"))
				exp = new TreeUnaryNode(autofree, pos, exp, OP_GREEN);
			else if (!strcmp(namebuf, "b"))
				exp = new TreeUnaryNode(autofree, pos, exp, OP_BLUE);
			else
				throw ParseError(getOffset(), "expected 'r', 'g', or 'b' after selection operator");
			break;

		default:
			return exp;
		}
	}
}

TreeExpressionNode *Parser::ParsePrefixExpression() {
	int pos = getOffset();

	switch(nexttoken()) {
	case '+':
		absorb();
		return ParsePrefixExpression();

	case '-':
		absorb();
		return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_NEG);

	case '~':
		absorb();
		return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_INV);

	case TOK_PLUSPLUS:
		absorb();
		return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_PREINC);

	case TOK_MINUSMINUS:
		absorb();
		return new TreeUnaryNode(autofree, pos, ParsePrefixExpression(), OP_PREDEC);

	default:
		return ParsePostfixExpression();
	}
}

TreeExpressionNode *Parser::ParseBinaryExpression2(TreeExpressionNode *e1, int precedence) {
	int tok;
	TreeExpressionNode *e2=NULL;
	Operator op;
	int pos = getOffset();

	while((tok = nexttoken()) && tok != ':' && tok != ')') {
		int prec2 = getPrecedence(tok);

		if (prec2 > precedence) {
			if (e2)
				e2 = ParseBinaryExpression2(e2, prec2);
			else
				e1 = ParseBinaryExpression2(e1, prec2);
		} else if (prec2 < precedence)
			break;
		else {
			absorb();

			if (e2)
				e1 = new TreeBinaryNode(autofree, pos, e1, e2, op);

			switch(tok) {
			case '+':			op = OP_ADD; break;
			case '-':			op = OP_SUB; break;
			case '*':			op = OP_MUL; break;
			case '/':			op = OP_DIV; break;
			case '&':			op = OP_AND; break;
			case '|':			op = OP_OR; break;
			case '^':			op = OP_XOR; break;
			case '<':			op = OP_CMPLT; break;
			case '>':			op = OP_CMPGT; break;
			case TOK_LE:		op = OP_CMPLE; break;
			case TOK_GE:		op = OP_CMPGE; break;
			case TOK_EQ:		op = OP_CMPEQ; break;
			case TOK_NE:		op = OP_CMPNE; break;
			case TOK_LAND:		op = OP_LAND; break;
			case TOK_LOR:		op = OP_LOR; break;
			case TOK_LSHIFT:	op = OP_LSHIFT; break;
			case TOK_RSHIFT:	op = OP_RSHIFT; break;
			case TOK_RLSHIFT:	op = OP_RLSHIFT; break;
			}

			e2 = ParsePrefixExpression();
		}
	}

	if (e2)
		e1 = new TreeBinaryNode(autofree, pos, e1, e2, op);

	return e1;
}

TreeExpressionNode *Parser::ParseBinaryExpression(int precedence) {
	TreeExpressionNode *p1, *p2;

	p1 = ParsePrefixExpression();
	p2 = ParseBinaryExpression2(p1, precedence);

	return p2;
}

TreeExpressionNode *Parser::ParseTernaryExpression(TreeExpressionNode *cond) {
	int tok;
	TreeExpressionNode *truep, *falsep;

	while((tok = nexttoken()) == '?') {
		absorb();
		truep = ParseExpression();
		expect(':');
		falsep = ParseTernaryExpression(ParseBinaryExpression(0));

		cond = new TreeTernaryNode(autofree, cond->offset, cond, truep, falsep);
	}

	return cond;
}

TreeExpressionNode *Parser::ParseExpression() {
	TreeExpressionNode *left = ParseBinaryExpression(0);

	if (nexttoken() == '=') {
		absorb();
		return new TreeAssignNode(autofree, left->offset, left, ParseExpression());
	} else
		return ParseTernaryExpression(left);
}

TreeStatementNode *Parser::ParseStatement() {
	int tok = nexttoken();
	int pos = getOffset();

	switch(tok) {
	case '{':
		{
			TreeStatementNode *pstFirst=NULL, *pstLast=NULL, *pst;
			absorb();
			while(nexttoken() != '}') {
				pst = ParseStatement();

				if (!pstFirst)
					pstFirst = pst;

				if (pstLast)
					pstLast->next = pst;

				pstLast = pst;
			}

			expect('}');

			return pstFirst;
		}

	case TOK_SCAN:
		{
			TreeScanNode *pScannerOld = pScanner;

			absorb();
			expect('(');
			TreeExpressionNode *bmpexp1 = ParseExpression(), *bmpexp2 = NULL;

			if (nexttoken() == ',') {
				absorb();

				bmpexp2 = ParseExpression();
			}

			expect(')');

			TreeScanNode *pScannerNew = new TreeScanNode(autofree, pos, bmpexp1, bmpexp2);
			
			pScanner = pScannerNew;
			pScannerNew->code = ParseStatement();
			pScanner = pScannerOld;

			return pScannerNew;
		}

	case TOK_IF:
		{
			absorb();
			expect('(');
			TreeExpressionNode *cond = ParseExpression();
			expect(')');
			TreeStatementNode *truep = ParseStatement(), *falsep = NULL;
			if (nexttoken() == TOK_ELSE) {
				absorb();
				falsep = ParseStatement();
			}

			return new TreeIfNode(autofree, pos, cond, truep, falsep);
		}

	case TOK_WHILE:
		{
			absorb();
			expect('(');
			TreeExpressionNode *cond = ParseExpression();
			expect(')');
			TreeStatementNode *path = ParseStatement();
			
			return new TreeWhileNode(autofree, pos, cond, path);
		}

	case TOK_DO:
		{
			absorb();
			TreeStatementNode *path = ParseStatement();
			expect(TOK_WHILE);
			expect('(');
			TreeExpressionNode *cond = ParseExpression();
			expect(')');
			expect(';');

			return new TreeDoWhileNode(autofree, pos, path, cond);
		}

	case TOK_FOR:
		{
			TreeExpressionNode *exp1 = NULL, *exp2 = NULL, *exp3 = NULL;
			TreeStatementNode *path, *exp3s;

			absorb();
			expect('(');
			if (nexttoken() != ';')
				exp1 = ParseExpression();
			expect(';');
			if (nexttoken() != ';')
				exp2 = ParseExpression();
			expect(';');
			if (nexttoken() != ')')
				exp3 = ParseExpression();
			expect(')');
			path = ParseStatement();

			if (exp3) {
				exp3s = new TreeExpStatementNode(autofree, pos, exp3);

				if (!path)
					path = exp3s;
				else {
					TreeStatementNode *pathend = path;

					while(pathend->next)
						pathend = pathend->next;

					pathend->next = exp3s;
				}
			}

			path = new TreeWhileNode(autofree, pos, exp2, path);

			if (exp1) {
				TreeExpStatementNode *exp1s = new TreeExpStatementNode(autofree, pos, exp1);

				exp1s->next = path;

				path = exp1s;
			}

			return path;
		}
		break;

	case TOK_RETURN:
		{
			TreeStatementNode *exp;

			absorb();

			if (pFunction->rettype != TYPE_VOID)
				exp = new TreeReturnNode(autofree, pos, ParseExpression());
			else
				exp = NULL;

			expect(';');

			return exp;
		}

	case TOK_INT:
	case TOK_BITMAP:
	case TOK_PIXEL32:
	case TOK_PIXEL64:
		{
			Type t = ParseType();
			int siz = getSize(t);

			for(;;) {
				expect(TOK_ID);

				TreeFunctionVariableNode *pNode = new TreeFunctionVariableNode(autofree, pos, namebuf, t, pFunction->localsize);

				pFunction->localsize += siz;

				pNode->next = pFunction->locals;
				pFunction->locals = pNode;

				if (nexttoken() != ',')
					break;
				absorb();
			}

			expect(';');
		}
		return NULL;
		

	case TOK_ID:
	case TOK_PIXVAL:
	case TOK_INTEGER:
	case TOK_IN:
	case TOK_OUT:
	case '(':
		{
			TreeStatementNode *path = new TreeExpStatementNode(autofree, pos, ParseExpression());
			expect(';');

			return path;
		}

	case ';':
		absorb();
		return NULL;

	default:
		throw ParseError(getOffset(), "expected statement");
	}
}

Type Parser::ParseType() {
	int tok = nexttoken();

	switch(tok) {
	case TOK_PIXEL32:
		absorb();
		return TYPE_PIXEL32;
	case TOK_PIXEL64:
		absorb();
		return TYPE_PIXEL64;
	case TOK_INT:
		absorb();
		return TYPE_INTEGER;
	case TOK_BITMAP:
		absorb();
		return TYPE_BITMAP;
	default:
		throw ParseError(getOffset(), "expected type (bitmap, pixel32, pixel64, or int)");
	}
}

TreeFunctionNode *Parser::ParseFunctionDefinition() {
	Type t;

	if (nexttoken() == TOK_VOID) {
		absorb();
		t = TYPE_VOID;
	} else
		t = ParseType();

	expect(TOK_ID);

	pFunction = new TreeFunctionNode(autofree, getOffset(), namebuf, t, NULL);
	TreeStatementNode *pStLast = NULL;

	expect('(');

	if (nexttoken() != ')') {
		for(;;) {
			Type argt = ParseType();

			expect(TOK_ID);

			pFunction->formalsize += getSize(argt);

			TreeFunctionVariableNode *pArg = new TreeFunctionVariableNode(autofree, getOffset(), namebuf, argt, -pFunction->formalsize-2);

			pArg->next = pFunction->formals;
			pFunction->formals = pArg;

			if (nexttoken() == ',')
				absorb();
			else
				break;
		}
	}

	expect(')');

	expect('{');

	while(nexttoken() != '}') {
		TreeStatementNode *pSt = ParseStatement();

		if (pSt) {
			if (!pStLast)
				pFunction->code = pSt;
			else
				pStLast->next = pSt;

			pStLast = pSt;
		}
	}

	expect('}');

	return pFunction;
}

TreeTranslationNode *Parser::ParseTranslationUnit() {
	TreeTranslationNode *ptn, *ptnFirst = NULL, *ptnLast;

	while(nexttoken()) {
		ptn = new TreeTranslationNode(autofree, getOffset(), ParseFunctionDefinition());

		if (!ptnFirst)
			ptnFirst = ptn;
		else
			ptnLast->next = ptn;

		ptnLast = ptn;
	}

	return ptnFirst;
}

///////////////////////////////////////////////////////////////////////////
//
//	external iface
//
///////////////////////////////////////////////////////////////////////////

Parser::Parser(ParseNodeAllocator *a) : autofree(a) {
}

Parser::~Parser() {
}

TreeTranslationNode *Parser::Parse(const char *str) {
	token_parsed = -1;
	tokptr = tokstart = tokptr_initial = str;
	pFunction = NULL;
	pScanner = NULL;

	return ParseTranslationUnit();
}
