//	filtfactory - poor man's filter compiler for VirtualDub
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_FILTFACTORY_PARSER_H
#define f_FILTFACTORY_PARSER_H

#include "list.h"

class TreeExpressionNode;
class TreeStatementNode;
class TreeFunctionNode;
class TreeFunctionVariableNode;
class TreeTranslationNode;
class TreeScanNode;

enum Type {
	TYPE_PIXEL32, TYPE_PIXEL64, TYPE_INTEGER, TYPE_VOID, TYPE_BITMAP,
	TYPE_PIXEL32PTR, TYPE_PIXEL64PTR, TYPE_INTPTR,
};

enum Operator {
	OP_ASSIGN,
	OP_ADD,
	OP_SUB,
	OP_MUL,
	OP_DIV,
	OP_CMPLT,
	OP_CMPGT,
	OP_CMPLE,
	OP_CMPGE,
	OP_CMPEQ,
	OP_CMPNE,
	OP_LAND,
	OP_LOR,
	OP_LSHIFT,
	OP_RSHIFT,
	OP_RLSHIFT,
	OP_AND,
	OP_OR,
	OP_XOR,
	OP_NEG,
	OP_INV,
	OP_PREINC,
	OP_PREDEC,
	OP_POSTINC,
	OP_POSTDEC,
	OP_RED,
	OP_GREEN,
	OP_BLUE,
	OP_CASTPIXEL32,
	OP_CASTPIXEL64,
};

enum Token {
	TOK_EOF		= 0,
	TOK_LE		= 256,
	TOK_GE,
	TOK_EQ,
	TOK_NE,
	TOK_PLUSPLUS,
	TOK_MINUSMINUS,
	TOK_LSHIFT,
	TOK_RSHIFT,
	TOK_RLSHIFT,
	TOK_LAND,
	TOK_LOR,
	TOK_ID,

	TOK_INTEGER,
	TOK_PIXVAL,

	TOK_PIXEL32,
	TOK_PIXEL64,
	TOK_INT,
	TOK_BITMAP,
	TOK_VOID,

	TOK_IN,
	TOK_OUT,

	TOK_SCAN,
	TOK_RETURN,
	TOK_IF,
	TOK_WHILE,
	TOK_DO,
	TOK_ELSE,
	TOK_FOR,
};

class ParseError {
private:
	struct ErrorString {
		int refcount;
		int offset;
		char estr[1];
	} *pError;

public:
	ParseError(int offset, const char *format, ...);
	ParseError(const ParseError&);
	~ParseError();

	ParseError& operator=(const ParseError&);

	const char *getError() {
		return pError ? pError->estr : "out of memory";
	}

	int getOffset() { return pError ? pError->offset : -1; }
};

class ParseNode;

class ParseNodeAllocator {
public:
	ListAlloc<ParseNode> listNodes;
};

class ParseNode : public ListNode2<ParseNode> {
private:
	ParseNodeAllocator *const a;

public:
	int const offset;

	ParseNode(ParseNodeAllocator *_a, int _offset) : a(_a), offset(_offset) {
		a->listNodes.AddTail(this);
	}

	void dispose() {
		Remove();
		delete this;
	}
	virtual ~ParseNode(){}
};

class Parser {
private:
	const char *tokptr_initial;
	const char *tokptr;
	const char *tokstart;
	int token_parsed;
	union {
		int i;
		unsigned long pix;
	} tokval;
	char namebuf[64];

	ParseNodeAllocator *const autofree;
	TreeFunctionNode *pFunction;
	TreeScanNode *pScanner;

	////////////////////////

	static int getPrecedence(int tok);
	static int getSize(Type t);
	static const char *getTokenName(char *buf, int tok);

	int getOffset() { return tokstart - tokptr_initial; }

	int		nexttoken();
	void	expect(int tok);
	void	absorb();

	TreeFunctionVariableNode *getVariable(const char *name);

	TreeExpressionNode *	ParseValue();
	TreeExpressionNode *	ParsePostfixExpression();
	TreeExpressionNode *	ParsePrefixExpression();
	TreeExpressionNode *	ParseBinaryExpression2(TreeExpressionNode *v, int precedence);
	TreeExpressionNode *	ParseBinaryExpression(int precedence);
	TreeExpressionNode *	ParseTernaryExpression(TreeExpressionNode *left);
	TreeExpressionNode *	ParseExpression();

	TreeStatementNode *		ParseStatement();
	Type	ParseType();
	TreeFunctionNode *		ParseFunctionDefinition();
	TreeTranslationNode *	ParseTranslationUnit();

public:
	Parser(ParseNodeAllocator *a);
	~Parser();

	TreeTranslationNode *	Parse(const char *str);
};

#endif
